#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define TYPE double
#include "../../../utils/blas.h"
#include "../../../utils/blas_init.h"
#include "../../../utils/timer.h"
#include "blocks/common/init.h"

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;


//
// Kernels
//

void dgemm_nobeta_tile(integer MBS, integer NBS, integer KBS, double ALPHA, double const A[MBS][KBS], double const B[KBS][NBS], double C[MBS][NBS]) {
	lapack_dgemm("N", "N", &NBS, &MBS, &KBS, &ALPHA, (double const *) B, &NBS, (double const *) A, &KBS, &dzero, (double *) C, &NBS);
}

void dgemm_tile(integer MBS, integer NBS, integer KBS, double ALPHA, double const A[MBS][KBS], double const B[KBS][NBS], double BETA, double C[MBS][NBS]) {
	lapack_dgemm("N", "N", &NBS, &MBS, &KBS, &ALPHA, (double const *) B, &NBS, (double const *) A, &KBS, &BETA, (double *) C, &NBS);
}

void tiled_dgemm(integer MBS, integer NBS, integer KBS, integer M, integer N, integer K, double ALPHA, double const A[M/MBS][K/KBS][MBS][KBS], double const B[K/KBS][N/NBS][KBS][NBS], double BETA, double C[M/MBS][N/NBS][MBS][NBS]) {
	#pragma omp parallel for
	for (integer i=0; i<M/MBS; i++) {
		for (integer j=0; j<N/NBS; j++) {
			#pragma omp task untied
			for (integer l=0; l<K/KBS; l++) {
				if (l==0) {
					if (BETA == 0.0) {
						dgemm_nobeta_tile(MBS, NBS, KBS, ALPHA, A[i][0], B[0][j], C[i][j]);
					} else {
						dgemm_tile(MBS, NBS, KBS, ALPHA, A[i][0], B[0][j], BETA, C[i][j]);
					}
				} else {
					dgemm_tile(MBS, NBS, KBS, ALPHA, A[i][l], B[l][j], 1.0, C[i][j]);
				}
			}
		}
	}
}

static inline double initializeElement(long I, long J, long i, long j) {
	static __thread uint64_t seed = 0;
	
	if ((i == 0) && (j == 0)) {
		seed = (I+i+1) * (J+j+1);
	}
	
	double result = (~0UL)/2UL;
	result -= seed;
	result /= ((~0UL / 2UL) + 1UL);
	seed = 6364136223846793005UL * seed + 1442695040888963407UL;
	
	return result;
}


static void idamax_task(integer N, const double X[N], double *result) {
	integer local_max = lapack_idamax(&N, X, &ONE);
	
	#pragma omp critical
	*result = max(*result, fabs(X[local_max-1]));
}


static void dasum_tile(integer MBS, integer NBS, integer M, integer N, const double X[MBS][NBS], double result[MBS]) {
	double local_result[MBS];
	
	for (integer j=0; j < MBS; j++)
		local_result[j] = 0.0;
	
	#pragma omp parallel for
	for (integer i=0; i < NBS; i++)
		for (integer j=0; j < MBS; j++)
			local_result[j] += fabs(X[j][i]);
	
	#pragma omp critical
	for (integer j=0; j < MBS; j++) {
		result[j] += local_result[j];
	}
}


static void substract(integer MBS, integer NBS, const double a[MBS][NBS], const double b[MBS][NBS], double c[MBS][NBS]) {
	integer N = MBS*NBS;
	
	lapack_dcopy(&N, (double const *) a, &ONE, (double *) c, &ONE);
	lapack_daxpy(&N, &dminusone, (double const *) b, &ONE, (double *) c, &ONE);
}


static void substract_iterative(integer MBS, integer NBS, integer M, integer N, const double a[M/MBS][N/NBS][MBS][NBS], const double b[M/MBS][N/NBS][MBS][NBS], double c[M/MBS][N/NBS][MBS][NBS]) {
	#pragma omp parallel for
	for (integer i = 0; i < M/MBS; i++) {
		for (integer j = 0; j < N/NBS; j++) {
			#pragma omp task untied
			substract(MBS, NBS, a[i][j], b[i][j], c[i][j]);
		}
	}
}


static void vector_norm_inf(integer BS, integer N, double const V[N/BS][BS], double *result) {
	*result = 0.0;
	
	#pragma omp parallel for
	for (long i=0; i < N/BS; i++)
		#pragma omp task untied
		idamax_task(BS, V[i], result);
}


static void matrix_norm_inf(integer MBS, integer NBS, integer M, integer N, double const A[M/MBS][N/NBS][MBS][NBS], double *result) {
	double (*sums)[MBS] = malloc( ((long)M) * sizeof(double) );
	
	#pragma omp parallel for
	for (long J=0; J < M/MBS; J++)
		for (long j=0; j < MBS; j++)
			sums[J][j] = 0.0;
	
	#pragma omp parallel for
	for (long i=0; i < N/NBS; i++)
		#pragma omp task untied
		for (long j=0; j < M/MBS; j++)
			dasum_tile(MBS, NBS, M, N, A[j][i], sums[j]);
	
	*result = 0.0;
	#pragma omp parallel for
	for (long j=0; j < M/MBS; j++)
		#pragma omp task untied
		idamax_task(MBS, sums[j], result);
	
	free(sums);
}


static void matrix_vector_product_nobeta(integer M, integer N, double A[M][N], double X[N], double Y[M]) {
	lapack_dgemv("T", &N, &M, &done, (double const *) A, &N, X, &ONE, &dzero, Y, &ONE);
}


static void matrix_vector_product(integer M, integer N, double A[M][N], double X[N], double BETA, double Y[M]) {
	lapack_dgemv("T", &N, &M, &done, (double const *) A, &N, X, &ONE, &BETA, Y, &ONE);
}


static void matrix_vector_product_iterative(integer MBS, integer NBS, integer M, integer N, double A[M/MBS][N/NBS][MBS][NBS], double X[N/NBS][NBS], double Y[M/MBS][MBS]) {
	#pragma omp parallel for
	for (integer i=0; i < M/MBS; i++) {
		#pragma omp task untied
		for (integer j=0; j < N/NBS; j++) {
			if (j == 0) {
				matrix_vector_product_nobeta(MBS, NBS, A[i][j], X[j], Y[i]);
			} else {
				matrix_vector_product(MBS, NBS, A[i][j], X[j], 1.0, Y[i]);
			}
		}
	}
}


static void dump_matrix(integer MBS, integer NBS, integer M, integer N, double A[M/MBS][N/NBS][MBS][NBS]) {
	for (integer I=0; I < M/MBS ; I++) {
		for (integer i=0; i < MBS; i++) {
			for (integer J=0; J < N/NBS; J++) {
				for (integer j=0; j < NBS; j++) {
					printf("%g\t", A[I][J][i][j]);
				}
			}
			printf("\n");
		}
	}
}


int main(int argc, char *argv[])
{
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <M> <N> <MBS> <NBS> <Index> <H|V|B|PW|SEQ>\n", argv[0]);
		exit(1);
	}
	
	unsigned long M = atol(argv[1]);
	unsigned long N = atol(argv[2]);
	unsigned long MBS = atol(argv[3]);
	unsigned long NBS = atol(argv[4]);
	unsigned long MaxBS = (MBS > NBS ? MBS : NBS);
	parseInitializationType(argv[6]);
	
	typedef double (*mn_p)[N/NBS][MBS][NBS];
	typedef double (*nn_p)[N/NBS][NBS][NBS];
	typedef double const (*mnc_p)[N/NBS][MBS][NBS];
	typedef double const (*nnc_p)[N/NBS][NBS][NBS];
	
	mn_p A = (mn_p) malloc(M * N * sizeof(double));
	nn_p B = (nn_p) malloc(N * N * sizeof(double));
	mn_p C = (mn_p) malloc(M * N * sizeof(double));
	
	initialize(M, N, MBS, NBS, A);
	initialize(N, N, NBS, NBS, B);
	force_blas_init();
	
	START_TIME();
	tiled_dgemm(MBS, NBS, NBS, M, N, N, 1.0, (mnc_p) A, (nnc_p) B, 0.0, C);
	STOP_TIME();
	
#ifndef NO_CHECK
	//
	// Verification inspired in the HPCC DGEMM verification
	//
	double (*X)[NBS] = malloc(1 * N * sizeof(double));
	double (*Y)[MBS] = malloc(1 * M * sizeof(double));
	double (*X2)[NBS] = malloc(1 * N * sizeof(double));
	double (*X3)[NBS] = malloc(1 * M * sizeof(double));
	double (*Z)[MBS] = malloc(1 * M * sizeof(double));
	
	double C_norm_inf;
	double X_norm_inf;
	
	initialize(1, N, 1, NBS, (double (*)[N/NBS][1][NBS]) X);
	
	#pragma omp parallel
	#pragma omp single
	{
		#pragma omp task untied
		matrix_norm_inf(MBS, NBS, M, N, (mnc_p) C, &C_norm_inf);
		#pragma omp task untied
		vector_norm_inf(NBS, N, X, &X_norm_inf);
		
		#pragma omp task untied
		matrix_vector_product_iterative(NBS, NBS, N, N, B, X, X2); // X2 := BX
		#pragma omp task untied
		matrix_vector_product_iterative(MBS, NBS, M, N, C, X, Z); // Z := CX
		
		#pragma omp taskwait
	}
	#pragma omp parallel
	#pragma omp single
	matrix_vector_product_iterative(MBS, NBS, M, N, A, X2, Y); // Y := A(X2)
	
	#pragma omp parallel
	#pragma omp single
	substract_iterative(1, MBS, // X3 := Y - Z === ABX - CX
		1, M,
		Y, Z, X3);
	
	double resid0;
	#pragma omp parallel
	#pragma omp single
	vector_norm_inf(MBS, M, X3, &resid0);
	
	free(X);
	free(X2);
	free(X3);
	free(Y);
	free(Z);
	
	double scaled_residual = resid0 / (C_norm_inf * X_norm_inf * N * DBL_EPSILON);
	
	if (scaled_residual >= 16.0) {
		fprintf(stderr, "Verification error: %f\n", scaled_residual);
		return 1;
	}
#endif
	
	if (getenv("MEASURE_TIME") == NULL) {
		// Performance in MFLOPS
		printf ("%f", ((2.0 * (double) M * (double) N * (double) N)/GET_TIME()) );
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	return 0;
}

