#ifndef USER_INITIALIZATION_BY_SHAPE_H
#define USER_INITIALIZATION_BY_SHAPE_H


#include <stdlib.h>
#include <string.h>


static enum {
	PARALLEL_LOOP_INIT,
	H_INIT,
	V_INIT,
	B_INIT,
	PW_INIT,
	SEQ_INIT
} initialization_type;


int initializationIsQuadratic() {
	return (initialization_type == B_INIT) || (initialization_type == PW_INIT);
}


#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
static inline TYPE initializeElement(long I, long J, long i, long j, long VN, long HN);
#else
static inline TYPE initializeElement(long I, long J, long i, long j);
#endif


static void zz_initializeBlock(long VN, long HN, long VBS, long HBS, long I, long J, TYPE data[VN][HN]) {
	for (long i=0; i < VBS; i++) {
		for (long j=0; j < HBS; j++) {
#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
			data[i][j] = initializeElement(I, J, i, j, VN, HN);
#else
			data[i][j] = initializeElement(I, J, i, j);
#endif
		}
	}
}


static void initialize(long VN, long HN, long VBS, long HBS, TYPE data[VN][HN]) {
	for (long I = 0; I < VN; I+=VBS) {
		#pragma omp parallel for
		for (long J = 0; J < HN; J+=HBS) {
			#pragma omp task
			{
				zz_initializeBlock(VN, HN, VBS, HBS, I, J, (TYPE (*)[HN]) &data[I][J]);
			}
		}
	}
}


static void zz_initializeBlockPair(long N, long BS, long I, long J, TYPE data1[N][N], TYPE data2[N][N]) {
	for (long i=0; i < BS; i++) {
		for (long j=0; j < BS; j++) {
#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
			data1[i][j] = initializeElement(I, J, i, j, N, N);
			data2[j][i] = initializeElement(J, I, j, i, N, N);
#else
			data1[i][j] = initializeElement(I, J, i, j);
			data2[j][i] = initializeElement(J, I, j, i);
#endif
		}
	}
}


static void initializeByPairs(long N, long BS, TYPE data[N][N]) {
	#pragma omp parallel for
	for (long I = 0; I < N; I+=BS) {
		for (long J = 0; J < I; J+=BS) {
			#pragma omp task
			{
				zz_initializeBlockPair(N, BS, I, J, (TYPE (*)[N]) &data[I][J], (TYPE (*)[N]) &data[J][I]);
			}
		}
		#pragma omp task
		{
			zz_initializeBlock(N, N, BS, BS, I, I, (TYPE (*)[N]) &data[I][I]);
		}
	}
}


static void initializeWithParallelLoops(long VN, long HN, TYPE data[VN][HN]) {
	#pragma omp parallel for
	for (long i=0; i < VN; i++) {
		for (long j=0; j < HN; j++) {
#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
			data[i][j] = initializeElement(0, 0, i, j, VN, HN);
#else
			data[i][j] = initializeElement(0, 0, i, j);
#endif
		}
	}
}


static void parseInitializationType(char const *initializazionTypeString) {
	if (strcmp(initializazionTypeString, "PL") == 0) {
		initialization_type = PARALLEL_LOOP_INIT;
	} else if (strcmp(initializazionTypeString, "I") == 0) {
		fprintf(stderr, "Error: Interleaving not available without sillymp.\n");
		exit(1);
	} else if (strcmp(initializazionTypeString, "H") == 0) {
		initialization_type = H_INIT;
	} else if (strcmp(initializazionTypeString, "V") == 0) {
		initialization_type = V_INIT;
	} else if (strcmp(initializazionTypeString, "B") == 0) {
		initialization_type = B_INIT;
	} else if (strcmp(initializazionTypeString, "PW") == 0) {
		initialization_type = PW_INIT;
	} else if (strcmp(initializazionTypeString, "SEQ") == 0) {
		initialization_type = SEQ_INIT;
	} else {
		fprintf(stderr, "Error: unknown initialization type parameter.\n");
		exit(1);
	}
}


static void initializeUsingUserSpecification(long VN, long HN, long VBS, long HBS, TYPE data[VN][HN]) {
	switch (initialization_type) {
		case PARALLEL_LOOP_INIT:
			// Initialize using OpenMP parallel loops
			initializeWithParallelLoops(VN, HN, data);
			break;
		case H_INIT:
			// Horizontal panels
			initialize(VN, HN, VBS, HN, data);
			break;
		case V_INIT:
			// Vertical panels
			initialize(VN, HN, VN, HBS, data);
			break;
		case B_INIT:
			// Blocks
			initialize(VN, HN, VBS, HBS, data);
			break;
		case PW_INIT:
			if (VN != HN) {
				fprintf(stderr, "Error: this problem does not allow pairwise initialization.\n");
				exit(1);
			}
			// Pairs of blocks around the diagonal
			initializeByPairs(VN, VBS, data);
			break;
		case SEQ_INIT:
			// Sequential
			initialize(VN, HN, VN, HN, data);
			break;
	}
}


#endif // USER_INITIALIZATION_BY_SHAPE_H
