#ifndef USER_INITIALIZATION_BY_SHAPE_H
#define USER_INITIALIZATION_BY_SHAPE_H


#include <stdlib.h>
#include <string.h>


static enum {
	PARALLEL_LOOP_INIT,
	H_INIT,
	V_INIT,
	B_INIT,
	PW_INIT,
	SEQ_INIT
} initialization_type;


static inline TYPE initializeElement(long I, long J, long i, long j);


static void zz_initializeBlock(long VN, long HN, long realVN, long realHN, long VBS, long HBS, long I, long J, TYPE data[realVN][realHN]) {
	for (long i=0; i < VBS; i++) {
		for (long j=0; j < HBS; j++) {
			data[i][j] = initializeElement(I, J, i, j);
		}
	}
}


static void initialize(long VN, long HN, long realVN, long realHN, long VBS, long HBS, TYPE data[realVN][realHN]) {
	for (long I = 0; I < VN; I+=VBS) {
		for (long J = 0; J < HN; J+=HBS) {
			#pragma omp task
			{
				zz_initializeBlock(VN, HN, realVN, realHN, VBS, HBS, I, J, (TYPE (*)[realHN]) &data[I][J]);
			}
		}
	}
	#pragma omp taskwait
}


static void zz_initializeBlockPair(long N, long realN, long BS, long I, long J, TYPE data1[realN][realN], TYPE data2[realN][realN]) {
	for (long i=0; i < BS; i++) {
		for (long j=0; j < BS; j++) {
			data1[i][j] = initializeElement(I, J, i, j);
			data2[j][i] = initializeElement(J, I, j, i);
		}
	}
}


static void initializeByPairs(long N, long realN, long BS, TYPE data[realN][realN]) {
	for (long I = 0; I < N; I+=BS) {
		for (long J = 0; J < I; J+=BS) {
			#pragma omp task
			{
				zz_initializeBlockPair(N, realN, BS, I, J, (TYPE (*)[realN]) &data[I][J], (TYPE (*)[realN]) &data[J][I]);
			}
		}
		#pragma omp task
		{
			zz_initializeBlock(N, N, realN, realN, BS, BS, I, I, (TYPE (*)[realN]) &data[I][I]);
		}
	}
	#pragma omp taskwait
}


static void initializeWithParallelLoops(long VN, long HN, long realVN, long realHN, TYPE data[realVN][realHN]) {
	#pragma omp parallel
	{
		#pragma omp for nowait
		for (long i=0; i < VN; i++) {
			for (long j=0; j < HN; j++) {
				data[i][j] = initializeElement(0, 0, i, j);
			}
		}
	}
}


static void parseInitializationType(char const *initializazionTypeString) {
	if (strcmp(initializazionTypeString, "PL") == 0) {
		initialization_type = PARALLEL_LOOP_INIT;
	} else if (strcmp(initializazionTypeString, "I") == 0) {
		fprintf(stderr, "Error: Interleaving not available without sillymp.\n");
		exit(1);
	} else if (strcmp(initializazionTypeString, "H") == 0) {
		initialization_type = H_INIT;
	} else if (strcmp(initializazionTypeString, "V") == 0) {
		initialization_type = V_INIT;
	} else if (strcmp(initializazionTypeString, "B") == 0) {
		initialization_type = B_INIT;
	} else if (strcmp(initializazionTypeString, "PW") == 0) {
		initialization_type = PW_INIT;
	} else if (strcmp(initializazionTypeString, "SEQ") == 0) {
		initialization_type = SEQ_INIT;
	} else {
		fprintf(stderr, "Error: unknown initialization type parameter.\n");
		exit(1);
	}
}


static void initializeUsingUserSpecification(long VN, long HN, long realVN, long realHN, long VBS, long HBS, TYPE data[realVN][realHN]) {
	switch (initialization_type) {
		case PARALLEL_LOOP_INIT:
			// Initialize using OpenMP parallel loops
			initializeWithParallelLoops(VN, HN, realVN, realHN, data);
			break;
		case H_INIT:
			// Horizontal panels
			#pragma omp parallel
			#pragma omp single
			initialize(VN, HN, realVN, realHN, VBS, HN, data);
			break;
		case V_INIT:
			// Vertical panels
			#pragma omp parallel
			#pragma omp single
			initialize(VN, HN, realVN, realHN, VN, HBS, data);
			break;
		case B_INIT:
			// Blocks
			#pragma omp parallel
			#pragma omp single
			initialize(VN, HN, realVN, realHN, VBS, HBS, data);
			break;
		case PW_INIT:
			if (VN != HN) {
				fprintf(stderr, "Error: this problem does not allow pairwise initialization.\n");
				exit(1);
			}
			// Pairs of blocks around the diagonal
			#pragma omp parallel
			#pragma omp single
			initializeByPairs(VN, realVN, VBS, data);
			break;
		case SEQ_INIT:
			// Sequential
			#pragma omp parallel
			#pragma omp single
			initialize(VN, HN, realVN, realHN, VN, HN, data);
			break;
	}
}


#endif // USER_INITIALIZATION_BY_SHAPE_H
