#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <assert.h>
#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
#define TYPE double
#include "../../../utils/blas.h"
#include "../../../utils/blas_init.h"
#include "../../../utils/timer.h"
#include "flat/common/init.h"

#define OVERRIDE_MALLOC_AND_FREE
#include "memory.h"

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;

static integer nthreads;


//
// Tasks of the main algorithm
//

static void dgetrf_tile(integer M, integer N, integer LDA, double A[N][LDA], integer const PIVoffset, integer IPIV[PIVoffset+min(M, N)]) {
	integer INFO;
	lapack_dgetrf(&M, &N, (double *) A, &LDA, &IPIV[PIVoffset], &INFO);
	
	for (int i=0; i < min(M, N); i++)
		IPIV[PIVoffset + i] += PIVoffset;
}

void dlaswp_tile(integer max_rows, integer N, integer LDA, double A[N][LDA], integer K1, integer K2, integer const PIVoffset, integer const IPIV[K2+1]) {
	K1 = K1 + PIVoffset + 1;
	K2 = K2 + PIVoffset + 1;
	lapack_dlaswp(&N, ((double *) &A[0][0]) - PIVoffset, &LDA, &K1, &K2, IPIV, &ONE);
}

static void dtrsm_tile(integer M, integer N, double ALPHA, integer LDA, double const A[M][LDA], integer LDB, double B[N][LDB]) {
	lapack_dtrsm("Left", "Lower", "No Transpose", "Unit", &M, &N, &ALPHA, (const double *) A, &LDA, (double *) B, &LDB);
}

static void dgemm_tile(integer M, integer N, integer K, double ALPHA, integer LDA, double const A[K][LDA], integer LDB, double const B[N][LDB], const double BETA, integer LDC, double C[N][LDC]) {
	lapack_dgemm("No Transpose", "No Transpose", &M, &N, &K, &ALPHA, (const double *) A, &LDA, (const double *) B, &LDB, &BETA, (double *) C, &LDC);
}

void add_scalar_tile(integer N, integer scalar, integer IPIV[N]) {
	for (int i=0; i < N; i++)
		IPIV[i] += scalar;
}


static void dtrsv_L_tile(integer NB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	lapack_dtrsv("L", "N", (check_diagonal ? "N" : "U"), &NB, (const double *) A, &N, X, &ONE);
}

static void dtrsv_U_tile(integer NB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	lapack_dtrsv("U", "N", (check_diagonal ? "N" : "U"), &NB, (double *) A, &N, X, &ONE);
}


static void pivot_vector(integer N, double v[N], integer const IPIV[N]) {
	for (int i=0; i < N; i++) {
		double tmp = v[i];
		v[i] = v[IPIV[i]-1];
		v[IPIV[i]-1] = tmp;
	}
}


static inline integer find_step_progression(int isFirst, integer remaining, integer smallStep, integer bigStep) {
	if (isFirst) {
		return smallStep;
	}
	
	while ((smallStep < bigStep) && (remaining >= smallStep*2L) && (remaining % (smallStep * 2L) == 0)) {
		smallStep = smallStep * 2L;
	}
	
	return smallStep;
}


//
// Decomposer functions
//

void tiled_scale(integer NB, integer N, integer scalar, integer IPIV[N]) {
	add_scalar_tile(N, scalar, IPIV);
}


void tiled_dlaswp(integer smallStep, integer bigStep, integer max_rows, integer N, integer LDA, double A[N][LDA], integer K1, integer K2, integer const PIVoffset, integer const IPIV[]) {
	typedef double (*m_p)[LDA];
	
	integer j_s = min(smallStep, N);
// 	#pragma omp for nowait
	for (int j=0; j < N; j += j_s) {
		#pragma omp task untied
		dlaswp_tile(max_rows, min(j_s, N-j), LDA, (m_p) &A[j][0], K1, K2, PIVoffset, IPIV);
	}
}


void tiled_dtrsm(integer smallStep, integer bigStep, integer M, integer N, double ALPHA, integer LDA, double const A[M][LDA], integer LDB, double B[N][LDB]) {
	typedef double (*b_p)[LDB];
	
	integer i_s = min(smallStep, N);
// 	#pragma omp for nowait
	for (integer i = 0; i < N; i += i_s) {
		#pragma omp task untied
		dtrsm_tile(M, min(i_s, N-i), ALPHA, LDA, A, LDB, (b_p) &B[i][0]);
	}
}


void tiled_dgemm(integer smallStep, integer bigStep, integer M, integer N, integer K, double ALPHA, integer LDA, double const A[K][LDA], integer LDB, double const B[N][LDB], const double BETA, integer LDC, double C[N][LDC]) {
	typedef const double (*a_p)[LDA];
	typedef const double (*b_p)[LDB];
	typedef double (*c_p)[LDC];
	
	// First block column made in blocks of smallStep for the M, N and K dimensions
	integer j_s = min(smallStep, N);
// 	#pragma omp for
	for (integer j = 0; j < N; j += j_s) {
		j_s = find_step_progression(j == 0, N - j, j_s, bigStep);
		for (integer l = 0; l < K; l += smallStep) {
			assert(K <= smallStep);
			integer l_s = min(smallStep, K-l); // K must always be smallStep. This makes l=0, that is, just one iteration.
			integer i_s = min(smallStep, M);
			for (integer i = 0; i < M; i += i_s) {
				i_s = find_step_progression(j == 0 /***/, M - i, i_s, bigStep); // The condition indicates that the small step is for the whole first column, which is controlled by "j"
				#pragma omp task untied
				dgemm_tile(i_s, j_s, l_s, ALPHA, LDA, (a_p) &A[l][i], LDB, (b_p) &B[j][l], BETA, LDC, (c_p) &C[j][i]);
			}
		}
	}
}


static inline integer round_to_power2(integer N) {
	if (N == 0) {
		return 0;
	} else if (N == 1) {
		return 1;
	} else if (N == 2) {
		return 2;
	} else if (N < 8) {
		return 4;
	} else if (N < 16) {
		return 8;
	} else if (N < 32) {
		return 16;
	} else if (N < 64) {
		return 32;
	} else if (N < 128) {
		return 64;
	} else if (N < 256) {
		return 128;
	} else if (N < 512) {
		return 256;
	} else if (N < 1024) {
		return 512;
	} else if (N < 2048) {
		return 1024;
	} else {
		return 1024 * round_to_power2(N / 1024);
	}
}


static inline integer calculate_small_step(integer total, integer want, integer best, integer minimum) {
	integer step = total / want;
	step = round_to_power2(step);
	
	if (step > best) {
		step = best;
	}
	if (step < minimum) {
		step = minimum;
	}
	
	return step;
}


static inline integer calculate_big_step(integer total, integer want, integer smallStep, integer best, integer minimum) {
	integer step = ((double) (total - smallStep - smallStep)) / sqrt(want);
	step = round_to_power2(step);
	
	if (step > best) {
		step = best;
	}
	if (step < minimum) {
		step = minimum;
	}
	
	return step;
}


void tiled_dgetrf(integer BESTNB, integer MINNB, integer M, integer N, integer LDA, double A[N][LDA], integer IPIV[], integer PIVoffset, integer RL) {
	typedef double (*a_p)[LDA];
	typedef const double (*ac_p)[LDA];
	
	integer SIZE = min(M, N);
	integer NB = calculate_small_step(SIZE, nthreads+1, BESTNB, MINNB);
	
	// Base case
	if (RL == 0 || NB == SIZE) {
		dgetrf_tile(M, N, LDA, A, PIVoffset, IPIV);
		return;
	}
	
	// Recursive case
	integer jb = NB;
	for (integer j=0; j < SIZE; j += jb) {
		// Adaptive block size calculation
		jb = calculate_small_step(SIZE-j, nthreads+1, BESTNB, MINNB);
		
		// Recursive call on diagonal block
		tiled_dgetrf(BESTNB, MINNB, M-j, jb, LDA, (a_p) &A[j][j], IPIV, PIVoffset+j, RL-1);
		
		// Adaptive block size calculation for the trailing matrix minus the first block column
		integer jb2 = calculate_big_step(SIZE-j, nthreads, jb, BESTNB, MINNB);
		
		#pragma omp taskwait
		if (j+jb < N) {
			// Row swaps on the right side
			tiled_dlaswp(jb, jb2, M, N-j-jb, LDA, (a_p) &A[j+jb][0], j, j+jb-1, PIVoffset, IPIV);
		}
		if (j != 0) {
			// Row swaps on the left side
			tiled_dlaswp(jb, jb, M, j, LDA, A, j, j+jb-1, PIVoffset, IPIV);
		}
		
		if (j+jb < N) {
			#pragma omp taskwait
			
			// Compute block row of U
			tiled_dtrsm(jb, jb2, jb, N-j-jb, 1.0, LDA, (ac_p) &A[j][j], LDA, (a_p) &A[j+jb][j]);
			#pragma omp taskwait
			
			if (j+jb < M) {
				// Update the trailing matrix
				tiled_dgemm(jb, jb2, M-j-jb, N-j-jb, jb, -1.0, LDA, (ac_p) &A[j][j+jb], LDA, (ac_p) &A[j+jb][j], 1.0, LDA, (a_p) &A[j+jb][j+jb]);
			}
		}
		
		#pragma omp taskwait
	}
}


//
// Tasks of the verification
//

static void dgemv_tile(integer M, integer N, double alpha, integer LDA, double const A[N][LDA], double const X[N], double beta, double Y[M]) {
	lapack_dgemv("N", &M, &N, &alpha, (const double *) A, &LDA, X, &ONE, &beta, Y, &ONE);
}


static void idamax_task(integer N, const double X[N], double *result) {
	integer local_max = lapack_idamax(&N, X, &ONE);
	
	#pragma omp critical
	*result = max(*result, fabs(X[local_max-1]));
}


static void dasum_tile(integer N, integer LDA, const double X[N][LDA], double result[N]) {
	double local_result[N];
	
	for (integer j=0; j < N; j++)
		local_result[j] = 0.0;
	
	for (integer i=0; i < N; i++)
		for (integer j=0; j < N; j++)
			local_result[j] += fabs(X[i][j]);
	
	for (integer j=0; j < N; j++) {
		#pragma omp critical
		result[j] += local_result[j];
	}
}


//
// Decomposer functions for the verification
//

static void smpss_dtrsv_U(integer BESTNB, integer MINNB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	typedef const double (*ac_p)[N];
	
// 	integer NB = calculate_small_step(N, nthreads+1, BESTNB, MINNB);
	integer NB = calculate_big_step(N, nthreads*2, 0, BESTNB, MINNB);
	for (integer i=N; i >= NB; i -= NB) {
// 		NB = calculate_small_step(i, nthreads+1, BESTNB, MINNB);
		dtrsv_U_tile(NB, N, (ac_p) &A[i-NB][i-NB], &X[i-NB], check_diagonal);
// 		#pragma omp for
		for (integer j=i-NB; j >= NB; j -= NB)
			#pragma omp task untied
			dgemv_tile(NB, NB, -1.0, N, (ac_p) &A[i-NB][j-NB], &X[i-NB], 1.0, &X[j-NB]);
		#pragma omp taskwait
	}
}


static void smpss_dtrsv_L(integer BESTNB, integer MINNB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	typedef const double (*ac_p)[N];
	
// 	integer NB = calculate_small_step(N, nthreads+1, BESTNB, MINNB);
	integer NB = calculate_big_step(N, nthreads*2, 0, BESTNB, MINNB);
	for (integer i=0; i < N; i += NB) {
// 		NB = calculate_small_step(N-i, nthreads+1, BESTNB, MINNB);
		dtrsv_L_tile(NB, N, (ac_p) &A[i][i], &X[i], check_diagonal);
// 		#pragma omp for
		for (integer j=i+NB; j < N; j += NB)
			#pragma omp task untied
			dgemv_tile(NB, NB, -1.0, N, (ac_p) &A[i][j], &X[i], 1.0, &X[j]);
		#pragma omp taskwait
	}
}


static void vector_norm_inf(integer NB, integer N, double const V[N], double *result) {
	*result = 0.0;
	
// 	#pragma omp for
	for (long i=0; i < N; i+=NB)
		#pragma omp task untied
		idamax_task(NB, &V[i], result);
}


static void matrix_norm_inf(integer NB, integer M, integer N, double const A[N][M], double *result) {
	typedef const double (*ac_p)[N];
	
	double *sums = malloc( ((long)M) * sizeof(double));
	
// 	#pragma omp for
	for (long j=0; j < M; j++)
		sums[j] = 0.0;
	
// 	#pragma omp for
	for (long i=0; i < N; i+=NB) {
		for (long j=0; j < M; j+=NB)
			#pragma omp task untied
			dasum_tile(NB, N, (ac_p) &A[i][j], &sums[j]);
		#pragma omp taskwait
	}
	
	*result = 0.0;
// 	#pragma omp for
	for (long j=0; j < M; j+=NB)
		#pragma omp task untied
		idamax_task(NB, &sums[j], result);
	
	#pragma omp taskwait
	
	free(sums);
}


//
// Main code
//

#if 1
#define RAND_SEED 100
#define RAND_MULT 6364136223846793005UL
#define RAND_ADD 1

static int64_t RAND_BS=0;
static int64_t RAND_MULT_BS=1;
static int64_t RAND_ADD_BS=0;
static int64_t RAND_N=0;
static int64_t RAND_MULT_N=1;
static int64_t RAND_ADD_N=0;

static inline void find_rand_jump_values(int64_t jump, int64_t *jump_mult, int64_t *jump_add) {
	for (; jump > 0UL; jump--) {
		*jump_mult = *jump_mult * RAND_MULT;
		*jump_add = *jump_add * RAND_MULT + RAND_ADD;
	}
}

static inline void advance_jump_values_by_subjump(
		int64_t *jump, int64_t *jump_mult, int64_t *jump_add,
		int64_t subjump, int64_t subjump_mult, int64_t subjump_add
) {
	for (; *jump >= subjump; *jump -= subjump) {
		*jump_mult = *jump_mult * subjump_mult;
		*jump_add = *jump_add * subjump_mult + subjump_add;
	}
}

static void initialize_precalculated_jump_values(long N, long BS) {
	RAND_N = N;
	RAND_BS = BS;
	RAND_MULT_BS=1;
	RAND_ADD_BS=0;
	RAND_MULT_N=1;
	RAND_ADD_N=0;
	
	find_rand_jump_values(BS, &RAND_MULT_BS, &RAND_ADD_BS);
	advance_jump_values_by_subjump(&N, &RAND_MULT_N, &RAND_ADD_N, BS, RAND_MULT_BS, RAND_ADD_BS);
	find_rand_jump_values(N, &RAND_MULT_N, &RAND_ADD_N);
}

static inline int64_t our_rand(int64_t position) {
	static __thread int64_t last_position = 0L;
	static __thread int64_t rand_state = RAND_SEED;
	
	if (position < last_position) {
		last_position = 0L;
		rand_state = RAND_SEED;
	}
	
	int64_t jump = position - last_position;
	int64_t jump_mult = 1;
	int64_t jump_add = 0;
	advance_jump_values_by_subjump(&jump, &jump_mult, &jump_add, RAND_N, RAND_MULT_N, RAND_ADD_N);
	advance_jump_values_by_subjump(&jump, &jump_mult, &jump_add, RAND_BS, RAND_MULT_BS, RAND_ADD_BS);
	find_rand_jump_values(jump, &jump_mult, &jump_add);
	
	rand_state = rand_state * jump_mult + jump_add;
	last_position = position;
	
	return rand_state;
}

static inline double initializeElement(long I, long J, long i, long j, long M, long N) {
	return 0.5 - ((double) our_rand((I+i)*N + J + j)) / ((double) (1UL << 63));
}	

#else

static void initialize_precalculated_jump_values(long N, long BS) {
}


static inline double initializeElement(long I, long J, long i, long j, long M, long N) {
	static __thread uint64_t seed = 0;
	
	if ((i == 0) && (j == 0)) {
		seed = (I+i+1) * (J+j+1);
	}
	
	double result = (~0UL)/2UL;
	result -= seed;
	result /= ((~0UL / 2UL) + 1UL);
	seed = 6364136223846793005UL * seed + 1442695040888963407UL;
	
	return result;
}
#endif


// NOTE: This function is called a second time for verification and must generate exactly the same sequence of random numbers
static void initialize_data(integer NB, integer N, double A[N][N], double B[N]) {
	initialize_precalculated_jump_values(N, NB);
	
	initializeUsingUserSpecification(N, N, NB, NB, A);
	
	srand(1);
	for (int i=0; i < N; i++)
		B[i] = ((double)RAND_MAX - 2.0*(double)rand()) / (double)RAND_MAX;
}


static void dump_matrix(integer M, integer N, double A[M][N]) {
	for (integer i=0; i < M ; i++) {
		for (integer j=0; j < N; j++) {
			printf("%g\t", A[i][j]);
		}
		printf("\n");
	}
}


static void dump_int_matrix(integer M, integer N, integer A[M][N]) {
	for (integer i=0; i < M ; i++) {
		for (integer j=0; j < N; j++) {
			printf("%ld\t", (integer)A[i][j]);
		}
		printf("\n");
	}
}


int main(int argc, char* argv[]) {
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <N> <N> <BestBS> <MinBS> <threads> <H|V|B|PW|SEQ>\n", argv[0]);
		return 1;
	}
	
	integer N = atol(argv[1]);
	integer BESTNB = atol(argv[3]);
	integer MINNB = atol(argv[4]);
	nthreads = atol(argv[5]);
	parseInitializationType(argv[6]);
	
	typedef double (*a_p)[N];
	typedef const double (*ac_p)[N];
	
	integer *IPIV = malloc(N*sizeof(integer)*2); // GotoBlas2 reads 2 elements after N
	a_p A = malloc(((long)N)*((long)N)*sizeof(double));
	double *B = malloc(((long)N)*sizeof(double));
	
	integer smallStep = calculate_small_step(N, nthreads, BESTNB, MINNB);
	integer bigStep = calculate_big_step(N, nthreads, smallStep, BESTNB, MINNB);
	
	if (initializationIsQuadratic()) {
		initialize_data(bigStep, N, A, B);
	} else {
		initialize_data(N/nthreads, N, A, B);
	}
	force_blas_init();
	
	START_TIME();
	#pragma omp parallel
	#pragma omp single
	{
		// 1. LU factorization
		tiled_dgetrf(BESTNB, MINNB, N, N, N, A, IPIV, 0, 2);
		
		// 2. Solve the system with the LU factorization
		// 2.1 Pivot vector B
		pivot_vector(N, B, IPIV);
		// 2.2 Forward elimination
		smpss_dtrsv_L(BESTNB, MINNB, N, (ac_p) A, B, 0);
		// 2.3 Backwards substitution
		smpss_dtrsv_U(BESTNB, MINNB, N, (ac_p) A, B, 1);
	}
	STOP_TIME();
	
	if (getenv("MEASURE_TIME") == NULL) {
		printf ("%f", (2.0*N*N*N/3.0+3.0*N*N/2.0)/GET_TIME());
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	double *X = B;
	B = malloc(((long)N)*sizeof(double));
	if (initializationIsQuadratic()) {
		initialize_data(bigStep, N, A, B);
	} else {
		initialize_data(N/nthreads, N, A, B);
	}
	#pragma omp taskwait
	
	double *v = malloc(((long)N)*sizeof(double));
	
	// v := B
	memcpy(v, B, ((long)N)*sizeof(double));
	
	// v := Ax - v
	lapack_dgemv("N", &N, &N, &done, (const double *)A, &N, X, &ONE, &dminusone, v, &ONE);
	
	double resid0;
	double A_norm_inf;
	double X_norm_inf;
	double B_norm_inf;
	
	#pragma omp parallel
	#pragma omp single
	{
		vector_norm_inf(bigStep, N, v, &resid0);
		
		matrix_norm_inf(bigStep, N, N, (ac_p) A, &A_norm_inf);
		vector_norm_inf(bigStep, N, X, &X_norm_inf);
		vector_norm_inf(bigStep, N, B, &B_norm_inf);
	}
	
	#pragma omp taskwait
	
	double resid1 = resid0 / (DBL_EPSILON * (A_norm_inf * X_norm_inf + B_norm_inf) * (double)N);
	
	if (resid1 >= 16.0) {
		fprintf(stderr, "Verification error: %f\n", resid1);
		return 1;
	}
	
	return 0;
}
