#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define TYPE double
#include "../../../utils/blas.h"
#include "../../../utils/blas_init.h"
#include "../../../utils/timer.h"
#include "flat/common/init.h"

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;


//
// Kernels
//

void dgemm_nobeta_tile(integer MBS, integer NBS, integer KBS, integer M, integer N, integer K, double ALPHA, double *A, double *B, double *C) {
	lapack_dgemm("N", "N", &NBS, &MBS, &KBS, &ALPHA, B, &N, A, &K, &dzero, C, &N);
}

void dgemm_tile(integer MBS, integer NBS, integer KBS, integer M, integer N, integer K, double ALPHA, double *A, double *B, double BETA, double *C) {
	lapack_dgemm("N", "N", &NBS, &MBS, &KBS, &ALPHA, B, &N, A, &K, &BETA, C, &N);
}

void tiled_dgemm(integer MBS, integer NBS, integer KBS, integer M, integer N, integer K, double ALPHA, double *A, double *B, double BETA, double *C) {
	#pragma omp parallel for
	for (integer i=0; i<M; i += MBS) {
		for (integer j=0; j<N; j+= NBS) {
			#pragma omp task untied
			for (integer l=0; l<K; l += KBS) {
				if (l==0) {
					if (BETA == 0.0) {
						dgemm_nobeta_tile(MBS, NBS, KBS, M, N, K, ALPHA, &A[l+i*K], &B[j+l*N], &C[j+i*N]);
					} else {
						dgemm_tile(MBS, NBS, KBS, M, N, K, ALPHA, &A[l+i*K], &B[j+l*N], BETA, &C[j+i*N]);
					}
				} else {
					dgemm_tile(MBS, NBS, KBS, M, N, K, ALPHA, &A[l+i*K], &B[j+l*N], 1.0, &C[j+i*N]);
				}
			}
		}
	}
}


static inline double initializeElement(long I, long J, long i, long j) {
	static __thread uint64_t seed = 0;
	
	if ((i == 0) && (j == 0)) {
		seed = (I+i+1) * (J+j+1);
	}
	
	double result = (~0UL)/2UL;
	result -= seed;
	result /= ((~0UL / 2UL) + 1UL);
	seed = 6364136223846793005UL * seed + 1442695040888963407UL;
	
	return result;
}


static void idamax_task(integer N, const double X[N], double *result) {
	integer local_max_index = lapack_idamax(&N, X, &ONE);
	double local_max = fabs(X[local_max_index-1]);
	
	#pragma omp critical
	*result = max(*result, local_max);
}


static void dasum_tile(integer M, integer N, integer LDA, const double X[M][LDA], double result[M]) {
	double local_result[M];
	
	for (integer j=0; j < M; j++)
		local_result[j] = 0.0;
	
	#pragma omp parallel for
	for (integer i=0; i < N; i++)
		for (integer j=0; j < M; j++)
			local_result[j] += fabs(X[j][i]);
	
	#pragma omp critical
	for (integer j=0; j < M; j++) {
		result[j] += local_result[j];
	}
}


static void substract(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer MBS, integer NBS) {
	for (integer i=0; i < MBS; i++) {
		lapack_dcopy(&NBS, a[i], &ONE, c[i], &ONE);
		lapack_daxpy(&NBS, &dminusone, b[i], &ONE, c[i], &ONE);
	}
}


static void substract_iterative(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, integer ML, integer NL, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer BS) {
	typedef double (*a_p)[AN];
	typedef double (*b_p)[BN];
	typedef double (*c_p)[CN];
	
	#pragma omp parallel for
	for (integer i = 0; i < ML; i+=BS) {
		for (integer j = 0; j < NL; j+=BS) {
			#pragma omp task untied
			substract(AM, AN, BM, BN, CM, CN, (a_p) &a[i][j], (b_p) &b[i][j], (c_p) &c[i][j], min(BS, ML-i), min(BS, NL-j));
		}
	}
}


static void vector_norm_inf(integer NB, integer N, double const V[N], double *result) {
	*result = 0.0;
	
	#pragma omp parallel for
	for (long i=0; i < N; i+=NB)
		#pragma omp task untied
		idamax_task(NB, &V[i], result);
}


static void matrix_norm_inf(integer MBS, integer NBS, integer M, integer N, double const A[M][N], double *result) {
	typedef double const (*m_p)[N];
	
	double *sums = malloc( ((long)M) * sizeof(double));
	
	#pragma omp parallel for
	for (long j=0; j < M; j++)
		sums[j] = 0.0;
	
	#pragma omp parallel for
	for (long i=0; i < N; i+=NBS)
		#pragma omp task untied
		for (long j=0; j < M; j+=MBS)
			dasum_tile(MBS, NBS, N, (m_p) &A[j][i], &sums[j]);
	
	*result = 0.0;
	#pragma omp parallel for
	for (long j=0; j < M; j+=MBS)
		#pragma omp task untied
		idamax_task(MBS, &sums[j], result);
	
	free(sums);
}


static void matrix_vector_product(integer M, integer N, double A[M][N], double X[N], double Y[M], integer BS) {
	lapack_dgemv("T", &N, &BS, &done, (double const *) A, &N, X, &ONE, &dzero, Y, &ONE);
}


static void matrix_vector_product_iterative(integer MBS, integer NBS, integer M, integer N, double A[M][N], double X[N], double Y[N]) {
	typedef double (*m_p)[N];
	
	#pragma omp parallel for
	for (integer i=0; i < M; i+=MBS) {
		#pragma omp task untied
		matrix_vector_product(M, N, (m_p) &A[i][0], X, &Y[i], MBS);
	}
}


static void dump_matrix(integer M, integer N, double A[M][N]) {
	for (integer i=0; i < M ; i++) {
		for (integer j=0; j < N; j++) {
			printf("%g\t", A[i][j]);
		}
		printf("\n");
	}
}


int main(int argc, char *argv[])
{
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <M> <N> <MBS> <NBS> <Index> <PL|H|V|B|PW|SEQ>\n", argv[0]);
		exit(1);
	}
	
	unsigned long M = atol(argv[1]);
	unsigned long N = atol(argv[2]);
	unsigned long MBS = atol(argv[3]);
	unsigned long NBS = atol(argv[4]);
	unsigned long MaxBS = (MBS > NBS ? MBS : NBS);
	parseInitializationType(argv[6]);
	
	typedef double (*m_p)[N];
	typedef double const (*mc_p)[N];
	
	double (*A)[N] = (m_p) malloc(M * N * sizeof(double));
	double (*B)[N] = (m_p) malloc(N * N * sizeof(double));
	double (*C)[N] = (m_p) malloc(M * N * sizeof(double));
	
	initializeUsingUserSpecification(M, N, MBS, NBS, A);
	initializeUsingUserSpecification(N, N, NBS, NBS, B);
	force_blas_init();
	
	START_TIME();
	tiled_dgemm(MBS, NBS, N, M, N, N, 1.0, (double *) A, (double *) B, 0.0, (double *) C);
	STOP_TIME();
	
#ifndef NO_CHECK
	//
	// Verification inspired in the HPCC DGEMM verification
	//
	double *X = malloc(1 * N * sizeof(double));
	double *Y = malloc(1 * M * sizeof(double));
	double *X2 = malloc(1 * N * sizeof(double));
	double *X3 = malloc(1 * M * sizeof(double));
	double *Z = malloc(1 * M * sizeof(double));
	
	double C_norm_inf;
	double X_norm_inf;
	
	typedef double (*n1_p)[N];
	typedef double const (*n1c_p)[N];
	typedef double (*m1_p)[M];
	typedef double const (*m1c_p)[M];
	
	initialize(1, N, 1, NBS, (n1_p) X);
	
	#pragma omp parallel
	#pragma omp single
	{
		#pragma omp task untied
		matrix_norm_inf(MBS, NBS, M, N, (mc_p) C, &C_norm_inf);
		#pragma omp task untied
		vector_norm_inf(NBS, N, X, &X_norm_inf);
		
		#pragma omp task untied
		matrix_vector_product_iterative(NBS, NBS, N, N, B, X, X2); // X2 := BX
		#pragma omp task untied
		matrix_vector_product_iterative(MBS, NBS, M, N, C, X, Z); // Z := CX
		
		#pragma omp taskwait
	}
	#pragma omp parallel
	#pragma omp single
	matrix_vector_product_iterative(MBS, NBS, M, N, A, X2, Y); // Y := A(X2)
	
	#pragma omp parallel
	#pragma omp single
	substract_iterative(1, M, 1, M, 1, M, // X3 := Y - Z === ABX - CX
		1, M,
		(m1_p) Y, (m1_p) Z, (m1_p) X, MBS);
	
	double resid0;
	#pragma omp parallel
	#pragma omp single
	vector_norm_inf(MBS, M, X3, &resid0);
	
	free(X);
	free(X2);
	free(X3);
	free(Y);
	free(Z);
	
	double scaled_residual = resid0 / (C_norm_inf * X_norm_inf * N * DBL_EPSILON);
	
	if (scaled_residual >= 16.0) {
		fprintf(stderr, "Verification error: %g\n", scaled_residual);
		return 1;
	}
#endif
	
	if (getenv("MEASURE_TIME") == NULL) {
		// Performance in MFLOPS
		printf ("%f", ((2.0 * (double) M * (double) N * (double) N)/GET_TIME()) );
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	return 0;
}

