#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define TYPE double
#include "../../../utils/blas.h"
#include "../../../utils/blas_init.h"
#include "../../../utils/timer.h"
#include "blocks/common/init.h"

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;



//
// Kernels
//

static void accumulate(integer MBS, integer NBS, double const a[MBS][NBS], double b[MBS][NBS]) {
	for (integer i=0; i < MBS; i++) {
		lapack_daxpy(&NBS, &done, a[i], &ONE, b[i], &ONE);
	}
}


static void add(integer MBS, integer NBS, double const a[MBS][NBS], double const b[MBS][NBS], double c[MBS][NBS]) {
	for (integer i=0; i < MBS; i++) {
		lapack_dcopy(&NBS, a[i], &ONE, c[i], &ONE);
		lapack_daxpy(&NBS, &done, b[i], &ONE, c[i], &ONE);
	}
}


static void substract(integer MBS, integer NBS, double const a[MBS][NBS], double const b[MBS][NBS], double c[MBS][NBS]) {
	integer N = MBS*NBS;
	
	lapack_dcopy(&N, (double const *) a, &ONE, (double *) c, &ONE);
	lapack_daxpy(&N, &dminusone, (double const *) b, &ONE, (double *) c, &ONE);
}


static void substract_and_accumulate(integer MBS, integer NBS, double const a[MBS][NBS], double const b[MBS][NBS], double c[MBS][NBS]) {
	integer N = MBS*NBS;
	
	lapack_daxpy(&N, &done, (double const *) a, &ONE, (double *) c, &ONE);
	lapack_daxpy(&N, &dminusone, (double const *) b, &ONE, (double *) c, &ONE);
}


void multiply_nobeta(integer MBS, integer NBS, integer KBS, double const A[MBS][KBS], double const B[KBS][NBS], double C[MBS][NBS]) {
	lapack_dgemm("N", "N", &NBS, &MBS, &KBS, &done, (double const *) B, &NBS, (double const *) A, &KBS, &dzero, (double *) C, &NBS);
}

void multiply(integer MBS, integer NBS, integer KBS, double const A[MBS][KBS], double const B[KBS][NBS], double BETA, double C[MBS][NBS]) {
	lapack_dgemm("N", "N", &NBS, &MBS, &KBS, &done, (double const *) B, &NBS, (double const *) A, &KBS, &BETA, (double *) C, &NBS);
}


static void invert_and_accumulate(integer MBS, integer NBS, double const a[MBS][NBS], double b[MBS][NBS]) {
	for (integer i=0; i < MBS; i++) {
		for (integer j=0; j < NBS; j++) {
			b[i][j] = a[i][j]-b[i][j];
		}
	}
}


//
// Iterative operations
//

static void accumulate_iterative(integer MBS, integer NBS, integer AM, integer AN, integer BM, integer BN, integer ML, integer NL, double const a[AM/MBS][AN/NBS][MBS][NBS], double b[BM/MBS][BN/NBS][MBS][NBS]) {
	for (integer i = 0; i < ML/MBS; i++) {
		for (integer j = 0; j < NL/NBS; j++) {
			#pragma omp task untied
			accumulate(MBS, NBS, a[i][j], b[i][j]);
		}
	}
	#pragma omp taskwait
}


static void add_iterative(integer MBS, integer NBS, integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, integer ML, integer NL, double const a[AM/MBS][AN/NBS][MBS][NBS], double const b[BM/MBS][BN/NBS][MBS][NBS], double c[CM/MBS][CN/NBS][MBS][NBS]) {
	for (integer i = 0; i < ML/MBS; i++) {
		for (integer j = 0; j < NL/NBS; j++) {
			#pragma omp task untied
			add(MBS, NBS, a[i][j], b[i][j], c[i][j]);
		}
	}
	#pragma omp taskwait
}


static void substract_iterative(integer MBS, integer NBS, integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, integer ML, integer NL, double const a[AM/MBS][AN/NBS][MBS][NBS], double const b[BM/MBS][BN/NBS][MBS][NBS], double c[CM/MBS][CN/NBS][MBS][NBS]) {
	for (integer i = 0; i < ML/MBS; i++) {
		for (integer j = 0; j < NL/NBS; j++) {
			#pragma omp task untied
			substract(MBS, NBS, a[i][j], b[i][j], c[i][j]);
		}
	}
	#pragma omp taskwait
}


static void substract_and_accumulate_iterative(integer MBS, integer NBS, integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, integer ML, integer NL, double const a[AM/MBS][AN/NBS][MBS][NBS], double const b[BM/MBS][BN/NBS][MBS][NBS], double c[CM/MBS][CN/NBS][MBS][NBS]) {
	for (integer i = 0; i < ML/MBS; i++) {
		for (integer j = 0; j < NL/NBS; j++) {
			#pragma omp task untied
			substract_and_accumulate(MBS, NBS, a[i][j], b[i][j], c[i][j]);
		}
	}
	#pragma omp taskwait
}


static void invert_and_accumulate_iterative(integer MBS, integer NBS, integer AM, integer AN, integer BM, integer BN, integer ML, integer NL, double const a[AM/MBS][AN/NBS][MBS][NBS], double b[BM/MBS][BN/NBS][MBS][NBS]) {
	for (integer i = 0; i < ML/MBS; i++) {
		for (integer j = 0; j < NL/NBS; j++) {
			#pragma omp task untied
			invert_and_accumulate(MBS, NBS, a[i][j], b[i][j]);
		}
	}
	#pragma omp taskwait
}


void multiply_iterative(integer MBS, integer NBS, integer KBS, integer AM, integer AK, integer BK, integer BN, integer CM, integer CN, integer ML, integer NL, integer KL, double const A[AM/MBS][AK/KBS][MBS][KBS], double const B[BK/KBS][BN/NBS][KBS][NBS], double C[CM/MBS][CN/NBS][MBS][NBS]) {
	for (integer i=0; i<ML/MBS; i ++) {
		for (integer j=0; j<NL/NBS; j++) {
			#pragma omp task untied
			for (integer l=0; l<KL/KBS; l++) {
				if (l==0) {
					multiply_nobeta(MBS, NBS, KBS, A[i][l], B[l][j], C[i][j]);
				} else {
					multiply(MBS, NBS, KBS, A[i][l], B[l][j], 1.0, C[i][j]);
				}
			}
		}
	}
	#pragma omp taskwait
}



void *alloc_hypermatrix(integer MBS, integer NBS, integer M, integer N) {
	return malloc(sizeof(double)*M*N);
}


void dealloc_hypermatrix(integer MBS, integer NBS, integer M, integer N, double (*a)[N/NBS][MBS][NBS]) {
	free(a);
}



//
// Recursive operations
//

// M, N and K determine the actual matrix sizes. ML, NL and KL determine their accessed regions.
// MBS, NBS and KBS are the dimensions of the inner blocking that correspond to each dimension.
static void strassen_multiply(integer MBS, integer NBS, integer KBS, integer AM, integer AK, integer BK, integer BN, integer CM, integer CN, integer ML, integer NL, integer KL, double const a[AM/MBS][AK/KBS][MBS][KBS], double const b[BK/KBS][BN/NBS][KBS][NBS], double c[CM/MBS][CN/NBS][MBS][NBS], integer MultBS) {
	if ((ML <= MultBS) || (NL <= MultBS) || (KL <= MultBS)) {
		multiply_iterative(MBS, NBS, KBS, AM, AK, BK, BN, CM, CN, ML, NL, KL, a, b, c);
	} else {
		typedef double const (*mkc_p)[AK/KBS][MBS][KBS];
		typedef double const (*knc_p)[BN/NBS][KBS][NBS];
		typedef double const (*mnc_p)[CN/NBS][MBS][NBS];
		typedef double (*mk_p)[AK/KBS][MBS][KBS];
		typedef double (*kn_p)[BN/NBS][KBS][NBS];
		typedef double (*mn_p)[CN/NBS][MBS][NBS];
		
		mkc_p a_11 = (mkc_p) &a[0][0];
		mkc_p a_12 = (mkc_p) &a[0][(KL/2)/KBS];
		mkc_p a_21 = (mkc_p) &a[(ML/2)/MBS][0];
		mkc_p a_22 = (mkc_p) &a[(ML/2)/MBS][(KL/2)/KBS];
		
		knc_p b_11 = (knc_p) &b[0][0];
		knc_p b_12 = (knc_p) &b[0][(NL/2)/NBS];
		knc_p b_21 = (knc_p) &b[(KL/2)/KBS][0];
		knc_p b_22 = (knc_p) &b[(KL/2)/KBS][(NL/2)/NBS];
		
		mn_p c_11 = (mn_p) &c[0][0];
		mn_p c_12 = (mn_p) &c[0][(NL/2)/NBS];
		mn_p c_21 = (mn_p) &c[(ML/2)/MBS][0];
		mn_p c_22 = (mn_p) &c[(ML/2)/MBS][(NL/2)/NBS];
		
		typedef double const (*sc_p)[(KL/2)/KBS][MBS][KBS];
		typedef double const (*tc_p)[(NL/2)/NBS][KBS][NBS];
		typedef double const (*pc_p)[(NL/2)/NBS][MBS][NBS];
		typedef double (*s_p)[(NL/2)/NBS][KBS][NBS];
		typedef double (*t_p)[(KL/2)/KBS][MBS][KBS];
		typedef double (*p_p)[(NL/2)/NBS][MBS][NBS];
		
		// Create storage for temporary calculations. Only HB*HB blocks are used.
		s_p s1 = (s_p) alloc_hypermatrix(MBS, KBS, ML/2, KL/2);
		s_p s2 = (s_p) alloc_hypermatrix(MBS, KBS, ML/2, KL/2);
		s_p s3 = (s_p) alloc_hypermatrix(MBS, KBS, ML/2, KL/2);
		s_p s4 = (s_p) alloc_hypermatrix(MBS, KBS, ML/2, KL/2);
		t_p t1 = (t_p) alloc_hypermatrix(KBS, NBS, KL/2, NL/2);
		t_p t2 = (t_p) alloc_hypermatrix(KBS, NBS, KL/2, NL/2);
		t_p t3 = (t_p) alloc_hypermatrix(KBS, NBS, KL/2, NL/2);
		t_p t4 = (t_p) alloc_hypermatrix(KBS, NBS, KL/2, NL/2);
		
		p_p p1 = (p_p) alloc_hypermatrix(MBS, NBS, ML/2, NL/2);
		p_p p6 = (p_p) alloc_hypermatrix(MBS, NBS, ML/2, NL/2);
		p_p p7 = (p_p) alloc_hypermatrix(MBS, NBS, ML/2, NL/2);
		
		p_p u4 = (p_p) alloc_hypermatrix(MBS, NBS, ML/2, NL/2);
		
		// 1. S1 := A21 + A22
		#pragma omp task untied
		add_iterative(MBS, KBS,
			AM, AK, AM, AK, ML/2, KL/2,
			ML/2, KL/2,
			a_21, a_22, s1);
		
		// 3. S3 := A11 - A21
		#pragma omp task untied
		substract_iterative(MBS, KBS,
			AM, AK, AM, AK, ML/2, KL/2,
			ML/2, KL/2,
			a_11, a_21, s3);
		
		// 4. T1 := B12 - B11
		#pragma omp task untied
		substract_iterative(KBS, NBS,
			BK, BN, BK, BN, KL/2, NL/2,
			KL/2, NL/2,
			b_12, b_11, t1);
		
		// 6. T3 := B22 - B12
		#pragma omp task untied
		substract_iterative(KBS, NBS,
			BK, BN, BK, BN, KL/2, NL/2,
			KL/2, NL/2,
			b_22, b_12, t3);
		
		// 9. P1 := A11 * B11
		#pragma omp task untied
		strassen_multiply(MBS, NBS, KBS,
			AM, AK, BK, BN, ML/2, NL/2,
			ML/2, NL/2, KL/2,
			a_11, b_11, p1,
			MultBS);
		
		// 10. P2 := A12 * B21 [-> C11]
		#pragma omp task untied
		strassen_multiply(MBS, NBS, KBS,
			AM, AK, BK, BN, CM, CN,
			ML/2, NL/2, KL/2,
			a_12, b_21, c_11,
			MultBS);
		
		#pragma omp taskwait
		
		// 2. S2 := S1 - A11
		#pragma omp task untied
		substract_iterative(MBS, KBS,
			ML/2, KL/2, AM, AK, ML/2, KL/2,
			ML/2, KL/2,
			(sc_p) s1, a_11, s2);
		
		// 5. T2 := B22 - T1
		#pragma omp task untied
		substract_iterative(KBS, NBS,
			BK, BN, KL/2, NL/2, KL/2, NL/2,
			KL/2, NL/2,
			b_22, (tc_p) t1, t2);
		
		// 13. P5 := S1 * T1 [-> C22]
		#pragma omp task untied
		strassen_multiply(MBS, NBS, KBS,
			ML/2, KL/2, KL/2, NL/2, CM, CN,
			ML/2, NL/2, KL/2,
			(sc_p) s1, (tc_p) t1, c_22,
			MultBS);
		
		// 15. P7 := S3 * T3
		#pragma omp task untied
		strassen_multiply(MBS, NBS, KBS,
			ML/2, KL/2, KL/2, NL/2, ML/2, NL/2,
			ML/2, NL/2, KL/2,
			(sc_p) s3, (tc_p) t3, p7,
			MultBS);
		
		// 16. C11 := P1 + P2[C11]
		#pragma omp task untied
		accumulate_iterative(MBS, NBS,
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			(pc_p) p1, c_11);
		
		#pragma omp taskwait
		
		// 7. S4 := A12 - S2
		#pragma omp task untied
		substract_iterative(MBS, KBS,
			AM, AK, ML/2, KL/2, ML/2, KL/2,
			ML/2, KL/2,
			a_12, (sc_p) s2, s4);
		
		// 8. T4 := T2 - B21
		#pragma omp task untied
		substract_iterative(KBS, NBS,
			KL/2, NL/2, BK, BN, KL/2, NL/2,
			KL/2, NL/2,
			(tc_p) t2, b_21, t4);
		
		// 14. P6 := S2 * T2
		#pragma omp task untied
		strassen_multiply(MBS, NBS, KBS,
			ML/2, KL/2, KL/2, NL/2, ML/2, NL/2,
			ML/2, NL/2, KL/2,
			(sc_p) s2, (tc_p) t2, p6,
			MultBS);
		
		#pragma omp taskwait
		
		// 11. P3 := S4 * B22 [-> C12]
		#pragma omp task untied
		strassen_multiply(MBS, NBS, KBS,
			ML/2, KL/2, BK, BN, CM, CN,
			ML/2, NL/2, KL/2,
			(sc_p) s4, b_22, c_12,
			MultBS);
		
		// 12. P4 := A22 * T4 [-> C21]
		#pragma omp task untied
		strassen_multiply(MBS, NBS, KBS,
			AM, AK, KL/2, NL/2, CM, CN,
			ML/2, NL/2, KL/2,
			a_22, (tc_p) t4, c_21,
			MultBS);
		
		// 17. U2 := P1 + P6 [-> P6]
		#pragma omp task untied
		accumulate_iterative(MBS, NBS,
			ML/2, NL/2, ML/2, NL/2,
			ML/2, NL/2,
			(pc_p) p1, p6);
		
		#pragma omp taskwait
		
		// 18. U3 := U2[P6] + P7 [-> P7]
		#pragma omp task untied
		accumulate_iterative(MBS, NBS,
			ML/2, NL/2, ML/2, NL/2,
			ML/2, NL/2,
			(pc_p) p6, p7);
		
		// 19. U4 := U2[P6] + P5[C22]
		#pragma omp task untied
		add_iterative(MBS, NBS,
			ML/2, NL/2, CM, CN, ML/2, NL/2,
			ML/2, NL/2,
			(pc_p) p6, (mnc_p) c_22, u4);
		
		#pragma omp taskwait
		
		// 20. C12 := U4 + P3[C12]
		#pragma omp task untied
		accumulate_iterative(MBS, NBS,
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			(pc_p) u4, c_12);
		
		// 21. C21 := U3[P7] - P4[C21]
		#pragma omp task untied
		invert_and_accumulate_iterative(MBS, NBS,
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			(pc_p) p7, c_21);
		
		// 22. C22 := U3[P7] + P5[C22]
		#pragma omp task untied
		accumulate_iterative(MBS, NBS,
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			(pc_p) p7, c_22);
		
		#pragma omp taskwait
		
		// Deallocate blocks for temporary calculations
		dealloc_hypermatrix(MBS, KBS, ML/2, KL/2, t1);
		dealloc_hypermatrix(MBS, KBS, ML/2, KL/2, t2);
		dealloc_hypermatrix(MBS, KBS, ML/2, KL/2, t3);
		dealloc_hypermatrix(MBS, KBS, ML/2, KL/2, t4);
		dealloc_hypermatrix(KBS, NBS, KL/2, NL/2, s1);
		dealloc_hypermatrix(KBS, NBS, KL/2, NL/2, s2);
		dealloc_hypermatrix(KBS, NBS, KL/2, NL/2, s3);
		dealloc_hypermatrix(KBS, NBS, KL/2, NL/2, s4);
		
		dealloc_hypermatrix(MBS, NBS, ML/2, NL/2, p1);
		dealloc_hypermatrix(MBS, NBS, ML/2, NL/2, p6);
		dealloc_hypermatrix(MBS, NBS, ML/2, NL/2, p7);
		
		dealloc_hypermatrix(MBS, NBS, ML/2, NL/2, u4);
	}
}


static inline double initializeElement(long I, long J, long i, long j) {
	static __thread uint64_t seed = 0;
	
	if ((i == 0) && (j == 0)) {
		seed = (I+i+1) * (J+j+1);
	}
	
	double result = (~0UL)/2UL;
	result -= seed;
	result /= ((~0UL / 2UL) + 1UL);
	seed = 6364136223846793005UL * seed + 1442695040888963407UL;
	
	return result;
}


static void idamax_task(integer N, const double X[N], double *result) {
	integer local_max_index = lapack_idamax(&N, X, &ONE);
	double local_max = fabs(X[local_max_index-1]);
	
	#pragma omp critical
	*result = max(*result, local_max);
}


static void dasum_tile(integer MBS, integer NBS, integer M, integer N, const double X[MBS][NBS], double result[MBS]) {
	double local_result[MBS];
	
	for (integer j=0; j < MBS; j++)
		local_result[j] = 0.0;
	
	for (integer i=0; i < NBS; i++)
		for (integer j=0; j < MBS; j++)
			local_result[j] += fabs(X[j][i]);
	
	for (integer j=0; j < MBS; j++) {
		result[j] += local_result[j];
	}
}


static void vector_norm_inf(integer BS, integer N, double const V[N/BS][BS], double *result) {
	*result = 0.0;
	
// 	#pragma omp for
	for (long i=0; i < N/BS; i++)
		#pragma omp task untied
		idamax_task(BS, V[i], result);
	#pragma omp taskwait
}


static void matrix_norm_inf(integer MBS, integer NBS, integer M, integer N, double const A[M/MBS][N/NBS][MBS][NBS], double *result) {
	double (*sums)[MBS] = malloc( ((long)M) * sizeof(double) );
	
// 	#pragma omp for
	for (long J=0; J < M/MBS; J++)
		#pragma omp task untied
		for (long j=0; j < MBS; j++)
			sums[J][j] = 0.0;
	#pragma omp taskwait
	
// 	#pragma omp for
	for (long i=0; i < N/NBS; i++)
		#pragma omp task untied
		for (long j=0; j < M/MBS; j++)
			dasum_tile(MBS, NBS, M, N, A[j][i], sums[j]);
	#pragma omp taskwait
	
	*result = 0.0;
// 	#pragma omp for
	for (long j=0; j < M/MBS; j++)
		#pragma omp task untied
		idamax_task(MBS, sums[j], result);
	#pragma omp taskwait
	
	free(sums);
}


static void matrix_vector_product_nobeta(integer M, integer N, double const A[M][N], double const X[N], double Y[M]) {
	lapack_dgemv("T", &N, &M, &done, (double const *) A, &N, X, &ONE, &dzero, Y, &ONE);
}


static void matrix_vector_product(integer M, integer N, double const A[M][N], double const X[N], double BETA, double Y[M]) {
	lapack_dgemv("T", &N, &M, &done, (double const *) A, &N, X, &ONE, &BETA, Y, &ONE);
}


static void matrix_vector_product_iterative(integer MBS, integer NBS, integer M, integer N, double const A[M/MBS][N/NBS][MBS][NBS], double const X[N/NBS][NBS], double Y[M/MBS][MBS]) {
// 	#pragma omp for
	for (integer i=0; i < M/MBS; i++) {
		#pragma omp task untied
		for (integer j=0; j < N/NBS; j++) {
			if (j == 0) {
				matrix_vector_product_nobeta(MBS, NBS, A[i][j], X[j], Y[i]);
			} else {
				matrix_vector_product(MBS, NBS, A[i][j], X[j], 1.0, Y[i]);
			}
		}
	}
	#pragma omp taskwait
}


static void dump_matrix(integer MBS, integer NBS, integer M, integer N, double A[M/MBS][N/NBS][MBS][NBS]) {
	for (integer I=0; I < M/MBS ; I++) {
		for (integer i=0; i < MBS; i++) {
			for (integer J=0; J < N/NBS; J++) {
				for (integer j=0; j < NBS; j++) {
					printf("%g\t", A[I][J][i][j]);
				}
			}
			printf("\n");
		}
	}
}


int main(int argc, char *argv[])
{
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <M> <N> <MultBS> <OtherBS> <Index> <PL|H|V|B|SEQ>\n", argv[0]);
		exit(1);
	}
	
	unsigned long M = atol(argv[1]);
	unsigned long N = atol(argv[2]);
	unsigned long K = N;
	unsigned long MultBS = atol(argv[3]);
	unsigned long OtherBS = atol(argv[4]);
	unsigned long MinBS = min(MultBS, OtherBS);
	parseInitializationType(argv[6]);
	
	integer MBS = MinBS;
	integer NBS = MinBS;
	integer KBS = MinBS;
	
	typedef double (*mn_p)[N/NBS][MBS][NBS];
	typedef double (*nn_p)[N/NBS][NBS][NBS];
	typedef double const (*mnc_p)[N/NBS][MBS][NBS];
	typedef double const (*nnc_p)[N/NBS][NBS][NBS];
	
	mn_p A = (mn_p) malloc(M * N * sizeof(double));
	nn_p B = (nn_p) malloc(N * N * sizeof(double));
	mn_p C = (mn_p) malloc(M * N * sizeof(double));
	
	initializeUsingUserSpecification(M, N, MinBS, MinBS, A);
	#pragma omp taskwait
	initializeUsingUserSpecification(N, N, MinBS, MinBS, B);
	
	force_blas_init();
	#pragma omp taskwait
	
	START_TIME();
	#pragma omp parallel
	#pragma omp single
	strassen_multiply(MBS, NBS, KBS, M, K, K, N, M, N, M, N, K, (mnc_p) A, (nnc_p) B, C, MultBS);
	STOP_TIME();
	
#ifndef NO_CHECK
	typedef double (*n1_p)[N/NBS][1][NBS];
	typedef double (*m1_p)[M/MBS][1][MBS];
	typedef double const (*n1c_p)[N/NBS][1][NBS];
	typedef double const (*m1c_p)[M/MBS][1][MBS];
	typedef double (*n_p)[NBS];
	typedef double const (*nc_p)[NBS];
	typedef double (*m_p)[MBS];
	typedef double const (*mc_p)[MBS];
	
	//
	// Verification inspired in the HPCC DGEMM verification
	//
	n_p X = malloc(1 * N * sizeof(double));
	m_p Y = malloc(1 * M * sizeof(double));
	n_p X2 = malloc(1 * N * sizeof(double));
	m_p X3 = malloc(1 * M * sizeof(double));
	
	m_p Z = malloc(1 * M * sizeof(double));
	
	double C_norm_inf;
	double X_norm_inf;
	
	initializeUsingUserSpecification(1, N, 1, NBS, (n1_p) X);
	
	#pragma omp parallel
	#pragma omp single
	{
		#pragma omp task untied
		matrix_norm_inf(MBS, NBS, M, N, (mnc_p) C, &C_norm_inf);
		#pragma omp task untied
		vector_norm_inf(NBS, N, (nc_p) X, &X_norm_inf);
		
		#pragma omp task untied
		matrix_vector_product_iterative(NBS, NBS, N, N, (nnc_p) B, (nc_p) X, X2); // X2 := BX
		#pragma omp task untied
		matrix_vector_product_iterative(MBS, NBS, M, N, (mnc_p) C, (nc_p) X, Z); // Z := CX
		
		#pragma omp taskwait
	}
	
	#pragma omp parallel
	#pragma omp single
	matrix_vector_product_iterative(MBS, NBS, M, N, (mnc_p) A, (nc_p) X2, Y); // Y := A(X2)
	
	#pragma omp parallel
	#pragma omp single
	substract_iterative(1, MBS, // X3 := Y - Z === ABX - CX
		1, M, 1, M, 1, M,
		1, M,
		(m1c_p) Y, (m1c_p) Z, (m1_p) X3);
	
	double resid0;
	#pragma omp parallel
	#pragma omp single
	vector_norm_inf(MBS, M, (mc_p) X3, &resid0);
	
	free(X);
	free(X2);
	free(X3);
	free(Y);
	free(Z);
	
	double scaled_residual = resid0 / (C_norm_inf * X_norm_inf * N * DBL_EPSILON);
	
	if (scaled_residual >= 16.0) {
		fprintf(stderr, "Verification error: %g\n", scaled_residual);
		return 1;
	}
#endif
	
	if (getenv("MEASURE_TIME") == NULL) {
		// Performance in MFLOPS of a standard matrix multiplication
		printf ("%f", ((2.0 * (double) M * (double) N * (double) K)/GET_TIME()) );
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	return 0;
}

