#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define TYPE double

#include "../../../utils/blas.h"
#include "../../../utils/blas_init.h"
#include "../../../utils/timer.h"
#include "flat/common/init.h"

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;


//
// Kernels
//

static void accumulate(integer AM, integer AN, integer BM, integer BN, double a[AM][AN], double b[BM][BN], integer MBS, integer NBS) {
	for (integer i=0; i < MBS; i++) {
		lapack_daxpy(&NBS, &done, a[i], &ONE, b[i], &ONE);
	}
}


static void add(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer MBS, integer NBS) {
	for (integer i=0; i < MBS; i++) {
		lapack_dcopy(&NBS, a[i], &ONE, c[i], &ONE);
		lapack_daxpy(&NBS, &done, b[i], &ONE, c[i], &ONE);
	}
}


static void substract(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer MBS, integer NBS) {
	for (integer i=0; i < MBS; i++) {
		lapack_dcopy(&NBS, a[i], &ONE, c[i], &ONE);
		lapack_daxpy(&NBS, &dminusone, b[i], &ONE, c[i], &ONE);
	}
}


static void substract_and_accumulate(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer MBS, integer NBS) {
	for (integer i=0; i < MBS; i++) {
		lapack_daxpy(&NBS, &done, a[i], &ONE, c[i], &ONE);
		lapack_daxpy(&NBS, &dminusone, b[i], &ONE, c[i], &ONE);
	}
}


static void multiply(integer AM, integer AK, integer BK, integer BN, integer CM, integer CN, integer ML, integer NL, integer KL, double a[AM][AK], double b[BK][BN], double c[CM][CN]) {
	lapack_dgemm("N", "N", &NL, &ML, &KL, &done, (double *)b, &BN, (double *)a, &AK, &dzero, (double *)c, &CN);
}


static void invert_and_accumulate(integer AM, integer AN, integer BM, integer BN, double a[AM][AN], double b[BM][BN], integer MBS, integer NBS) {
	for (integer i=0; i < MBS; i++) {
		for (integer j=0; j < NBS; j++) {
			b[i][j] = a[i][j]-b[i][j];
		}
	}
}


//
// Iterative operations
//

static void accumulate_iterative(integer AM, integer AN, integer BM, integer BN, integer ML, integer NL, double a[AM][AN], double b[BM][BN], integer BS) {
	typedef double (*a_p)[AN];
	typedef double (*b_p)[BN];
	
	for (integer i = 0; i < ML; i+=BS) {
		for (integer j = 0; j < NL; j+=BS) {
			#pragma omp task untied
			accumulate(AM, AN, BM, BN, (a_p) &a[i][j], (b_p) &b[i][j], min(BS, ML-i), min(BS, NL-j));
		}
	}
	#pragma omp taskwait
}


static void add_iterative(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, integer ML, integer NL, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer BS) {
	typedef double (*a_p)[AN];
	typedef double (*b_p)[BN];
	typedef double (*c_p)[CN];
	
	for (integer i = 0; i < ML; i+=BS) {
		for (integer j = 0; j < NL; j+=BS) {
			#pragma omp task untied
			add(AM, AN, BM, BN, CM, CN, (a_p) &a[i][j], (b_p) &b[i][j], (c_p) &c[i][j], min(BS, ML-i), min(BS, NL-j));
		}
	}
	#pragma omp taskwait
}


static void substract_iterative(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, integer ML, integer NL, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer BS) {
	typedef double (*a_p)[AN];
	typedef double (*b_p)[BN];
	typedef double (*c_p)[CN];
	
	for (integer i = 0; i < ML; i+=BS) {
		for (integer j = 0; j < NL; j+=BS) {
			#pragma omp task untied
			substract(AM, AN, BM, BN, CM, CN, (a_p) &a[i][j], (b_p) &b[i][j], (c_p) &c[i][j], min(BS, ML-i), min(BS, NL-j));
		}
	}
	#pragma omp taskwait
}


static void substract_and_accumulate_iterative(integer AM, integer AN, integer BM, integer BN, integer CM, integer CN, integer ML, integer NL, double a[AM][AN], double b[BM][BN], double c[CM][CN], integer BS) {
	typedef double (*a_p)[AN];
	typedef double (*b_p)[BN];
	typedef double (*c_p)[CN];
	
	for (integer i = 0; i < ML; i+=BS) {
		for (integer j = 0; j < NL; j+=BS) {
			#pragma omp task untied
			substract_and_accumulate(AM, AN, BM, BN, CM, CN, (a_p) &a[i][j], (b_p) &b[i][j], (c_p) &c[i][j], min(BS, ML-i), min(BS, NL-j));
		}
	}
	#pragma omp taskwait
}


static void invert_and_accumulate_iterative(integer AM, integer AN, integer BM, integer BN, integer ML, integer NL, double a[AM][AN], double b[BM][BN], integer BS) {
	typedef double (*a_p)[AN];
	typedef double (*b_p)[BN];
	
	for (integer i = 0; i < ML; i+=BS) {
		for (integer j = 0; j < NL; j+=BS) {
			#pragma omp task untied
			invert_and_accumulate(AM, AN, BM, BN, (a_p) &a[i][j], (b_p) &b[i][j], min(BS, ML-i), min(BS, NL-j));
		}
	}
	#pragma omp taskwait
}





//
// Recursive operations
//

// M, N and K determine the actual matrix sizes. ML, NL and KL determine their accessed regions.
// The actual regions are:
// 	a[0;ML][0;KL]
// 	b[0;KL][0;NL]
// 	c[0;ML][0;NL]
static void strassen_multiply(integer AM, integer AK, integer BK, integer BN, integer CM, integer CN, integer ML, integer NL, integer KL, double a[AM][AK], double b[BK][BN], double c[CM][CN], integer MultBS, integer OtherBS) {
	if ((ML <= MultBS) || (NL <= MultBS) || (KL <= MultBS)) {
		multiply(AM, AK, BK, BN, CM, CN, ML, NL, KL, a, b, c);
	} else {
		// Since we pass t1 and t2 recursively as a and b, we need to make sure we do not overwrite their
		// original data when calculating M1-M6, since the data in a_11 and b_11 (which would be overwriten
		// since they coincide with t1 and t2) is read by M2-M6.
		double (*a_11)[AK] = (double (*)[AK]) &a[0][0];
		double (*a_12)[AK] = (double (*)[AK]) &a[0][KL/2];
		double (*a_21)[AK] = (double (*)[AK]) &a[ML/2][0];
		double (*a_22)[AK] = (double (*)[AK]) &a[ML/2][KL/2];
		
		double (*b_11)[BN] = (double (*)[BN]) &b[0][0];
		double (*b_12)[BN] = (double (*)[BN]) &b[0][NL/2];
		double (*b_21)[BN] = (double (*)[BN]) &b[KL/2][0];
		double (*b_22)[BN] = (double (*)[BN]) &b[KL/2][NL/2];
		
		double (*c_11)[CN] = (double (*)[CN]) &c[0][0];
		double (*c_12)[CN] = (double (*)[CN]) &c[0][NL/2];
		double (*c_21)[CN] = (double (*)[CN]) &c[ML/2][0];
		double (*c_22)[CN] = (double (*)[CN]) &c[ML/2][NL/2];
		
		// Create storage for temporary calculations. Only HB*HB blocks are used.
		double (*s1)[ML/2] = (double (*)[ML/2]) malloc(sizeof(double)*(ML*KL)/4);
		double (*s2)[ML/2] = (double (*)[ML/2]) malloc(sizeof(double)*(ML*KL)/4);
		double (*s3)[ML/2] = (double (*)[ML/2]) malloc(sizeof(double)*(ML*KL)/4);
		double (*s4)[ML/2] = (double (*)[ML/2]) malloc(sizeof(double)*(ML*KL)/4);
		double (*t1)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(KL*NL)/4);
		double (*t2)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(KL*NL)/4);
		double (*t3)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(KL*NL)/4);
		double (*t4)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(KL*NL)/4);
		
		double (*p1)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(ML*NL)/4);
		double (*p6)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(ML*NL)/4);
		double (*p7)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(ML*NL)/4);
		
		double (*u4)[NL/2] = (double (*)[NL/2]) malloc(sizeof(double)*(ML*NL)/4);
		
		// 1. S1 := A21 + A22
		#pragma omp task untied
		add_iterative(
			AM, AK, AM, AK, ML/2, KL/2,
			ML/2, KL/2,
			a_21, a_22, s1, OtherBS);
		
		// 3. S3 := A11 - A21
		#pragma omp task untied
		substract_iterative(
			AM, AK, AM, AK, ML/2, KL/2,
			ML/2, KL/2,
			a_11, a_21, s3, OtherBS);
		
		// 4. T1 := B12 - B11
		#pragma omp task untied
		substract_iterative(
			BK, BN, BK, BN, KL/2, NL/2,
			KL/2, NL/2,
			b_12, b_11, t1, OtherBS);
		
		// 6. T3 := B22 - B12
		#pragma omp task untied
		substract_iterative(
			BK, BN, BK, BN, KL/2, NL/2,
			KL/2, NL/2,
			b_22, b_12, t3, OtherBS);
		
		// 9. P1 := A11 * B11
		#pragma omp task untied
		strassen_multiply(
			AM, AK, BK, BN, ML/2, NL/2,
			ML/2, NL/2, KL/2,
			a_11, b_11, p1,
			MultBS, OtherBS);
		
		// 10. P2 := A12 * B21 [-> C11]
		#pragma omp task untied
		strassen_multiply(
			AM, AK, BK, BN, CM, CN,
			ML/2, NL/2, KL/2,
			a_12, b_21, c_11,
			MultBS, OtherBS);
		
		#pragma omp taskwait
		
		// 2. S2 := S1 - A11
		#pragma omp task untied
		substract_iterative(
			ML/2, KL/2, AM, AK, ML/2, KL/2,
			ML/2, KL/2,
			s1, a_11, s2, OtherBS);
		
		// 5. T2 := B22 - T1
		#pragma omp task untied
		substract_iterative(
			BK, BN, KL/2, NL/2, KL/2, NL/2,
			KL/2, NL/2,
			b_22, t1, t2, OtherBS);
		
		// 13. P5 := S1 * T1 [-> C22]
		#pragma omp task untied
		strassen_multiply(
			ML/2, KL/2, KL/2, NL/2, CM, CN,
			ML/2, NL/2, KL/2,
			s1, t1, c_22,
			MultBS, OtherBS);
		
		// 15. P7 := S3 * T3
		#pragma omp task untied
		strassen_multiply(
			ML/2, KL/2, KL/2, NL/2, ML/2, NL/2,
			ML/2, NL/2, KL/2,
			s3, t3, p7,
			MultBS, OtherBS);
		
		// 16. C11 := P1 + P2[C11]
		#pragma omp task untied
		accumulate_iterative(
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			p1, c_11, OtherBS);
		
		#pragma omp taskwait
		
		// 7. S4 := A12 - S2
		#pragma omp task untied
		substract_iterative(
			AM, AK, ML/2, KL/2, ML/2, KL/2,
			ML/2, KL/2,
			a_12, s2, s4, OtherBS);
		
		// 8. T4 := T2 - B21
		#pragma omp task untied
		substract_iterative(
			KL/2, NL/2, BK, BN, KL/2, NL/2,
			KL/2, NL/2,
			t2, b_21, t4, OtherBS);
		
		// 14. P6 := S2 * T2
		#pragma omp task untied
		strassen_multiply(
			ML/2, KL/2, KL/2, NL/2, ML/2, NL/2,
			ML/2, NL/2, KL/2,
			s2, t2, p6,
			MultBS, OtherBS);
		
		#pragma omp taskwait
		
		// 11. P3 := S4 * B22 [-> C12]
		#pragma omp task untied
		strassen_multiply(
			ML/2, KL/2, BK, BN, CM, CN,
			ML/2, NL/2, KL/2,
			s4, b_22, c_12,
			MultBS, OtherBS);
		
		// 12. P4 := A22 * T4 [-> C21]
		#pragma omp task untied
		strassen_multiply(
			AM, AK, KL/2, NL/2, CM, CN,
			ML/2, NL/2, KL/2,
			a_22, t4, c_21,
			MultBS, OtherBS);
		
		// 17. U2 := P1 + P6 [-> P6]
		#pragma omp task untied
		accumulate_iterative(
			ML/2, NL/2, ML/2, NL/2,
			ML/2, NL/2,
			p1, p6, OtherBS);
		
		#pragma omp taskwait
		
		// 18. U3 := U2[P6] + P7 [-> P7]
		#pragma omp task untied
		accumulate_iterative(
			ML/2, NL/2, ML/2, NL/2,
			ML/2, NL/2,
			p6, p7, OtherBS);
		
		// 19. U4 := U2[P6] + P5[C22]
		#pragma omp task untied
		add_iterative(
			ML/2, NL/2, CM, CN, ML/2, NL/2,
			ML/2, NL/2,
			p6, c_22, u4, OtherBS);
		
		#pragma omp taskwait
		
		// 20. C12 := U4 + P3[C12]
		#pragma omp task untied
		accumulate_iterative(
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			u4, c_12, OtherBS);
		
		// 21. C21 := U3[P7] - P4[C21]
		#pragma omp task untied
		invert_and_accumulate_iterative(
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			p7, c_21, OtherBS);
		
		// 22. C22 := U3[P7] + P5[C22]
		#pragma omp task untied
		accumulate_iterative(
			ML/2, NL/2, CM, CN,
			ML/2, NL/2,
			p7, c_22, OtherBS);
		
		#pragma omp taskwait
		
		// Deallocate blocks for temporary calculations
		free(t1);
		free(t2);
		free(t3);
		free(t4);
		free(s1);
		free(s2);
		free(s3);
		free(s4);
		
		free(p1);
		free(p6);
		free(p7);
		
		free(u4);
	}
}


static inline double initializeElement(long I, long J, long i, long j) {
	static __thread uint64_t seed = 0;
	
	if ((i == 0) && (j == 0)) {
		seed = (I+i+1) * (J+j+1);
	}
	
	double result = (~0UL)/2UL;
	result -= seed;
	result /= ((~0UL / 2UL) + 1UL);
	seed = 6364136223846793005UL * seed + 1442695040888963407UL;
	
	return result;
}


static void idamax_task(integer N, const double X[N], double *result) {
	integer local_max = lapack_idamax(&N, X, &ONE);
	
	#pragma omp critical
	*result = max(*result, fabs(X[local_max-1]));
}


static void dasum_tile(integer M, integer N, integer LDA, const double X[M][LDA], double result[M]) {
	double local_result[M];
	
	for (integer j=0; j < M; j++)
		local_result[j] = 0.0;
	
	for (integer i=0; i < N; i++)
		for (integer j=0; j < M; j++)
			local_result[j] += fabs(X[j][i]);
	
	for (integer j=0; j < M; j++) {
		#pragma omp critical
		result[j] += local_result[j];
	}
}


static void vector_norm_inf(integer BS, integer N, double const V[N], double *result) {
	*result = 0.0;
	
// 	#pragma omp for
	for (long i=0; i < N; i+=BS)
		#pragma omp task untied
		idamax_task(BS, &V[i], result);
	#pragma omp taskwait
}


static void matrix_norm_inf(integer MBS, integer NBS, integer M, integer N, double const A[M][N], double *result) {
	typedef double const (*m_p)[N];
	
	double *sums = malloc( ((long)M) * sizeof(double));
	
// 	#pragma omp for
	for (long j=0; j < M; j++)
		sums[j] = 0.0;
	
// 	#pragma omp for
	for (long i=0; i < N; i+=NBS)
		#pragma omp task untied
		for (long j=0; j < M; j+=MBS)
			dasum_tile(MBS, NBS, N, (m_p) &A[j][i], &sums[j]);
	#pragma omp taskwait
	
	*result = 0.0;
// 	#pragma omp for
	for (long j=0; j < M; j+=MBS)
		#pragma omp task untied
		idamax_task(MBS, &sums[j], result);
	#pragma omp taskwait
	
	free(sums);
}


static void matrix_vector_product(integer M, integer N, double A[M][N], double X[N], double Y[M], integer BS) {
	lapack_dgemv("T", &N, &BS, &done, (double const *) A, &N, X, &ONE, &dzero, Y, &ONE);
}


static void matrix_vector_product_iterative(integer MBS, integer NBS, integer M, integer N, double A[M][N], double X[N], double Y[N]) {
	typedef double (*m_p)[N];
	
// 	#pragma omp for
	for (integer i=0; i < M; i+=MBS) {
		#pragma omp task untied
		matrix_vector_product(M, N, (m_p) &A[i][0], X, &Y[i], MBS);
	}
	#pragma omp taskwait
}


static void dump_matrix(integer M, integer N, double A[M][N]) {
	for (integer i=0; i < M ; i++) {
		for (integer j=0; j < N; j++) {
			printf("%g\t", A[i][j]);
		}
		printf("\n");
	}
}


int main(int argc, char *argv[])
{
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <M> <N> <MultBS> <OtherBS> <Index> <H|V|B|PW|SEQ|X>\n", argv[0]);
		exit(1);
	}
	
	unsigned long M = atol(argv[1]);
	unsigned long N = atol(argv[2]);
	unsigned long K = N;
	unsigned long MultBS = atol(argv[3]);
	unsigned long OtherBS = atol(argv[4]);
	unsigned long MaxBS = min( max(MultBS, OtherBS), min(M, N) );
	parseInitializationType(argv[6]);
	
	typedef double (*m_p)[N];
	typedef double const (*mc_p)[N];
	
	double (*A)[N] = (m_p) malloc(M * N * sizeof(double));
	double (*B)[N] = (m_p) malloc(N * N * sizeof(double));
	double (*C)[N] = (m_p) malloc(M * N * sizeof(double));
	
	initializeUsingUserSpecification(M, N, MaxBS, MaxBS, A);
	#pragma omp taskwait
	initializeUsingUserSpecification(N, N, MaxBS, MaxBS, B);
	
	force_blas_init();
	#pragma omp taskwait
	
	START_TIME();
	#pragma omp parallel
	#pragma omp single
	strassen_multiply(M, K, K, N, M, N, M, N, K, A, B, C, MultBS, OtherBS);
	STOP_TIME();
	
#ifndef NO_CHECK
	//
	// Verification inspired in the HPCC DGEMM verification
	//
	double *X = malloc(1 * N * sizeof(double));
	double *Y = malloc(1 * M * sizeof(double));
	double *X2 = malloc(1 * N * sizeof(double));
	double *X3 = malloc(1 * M * sizeof(double));
	
	double *Z = malloc(1 * M * sizeof(double));
	
	double C_norm_inf;
	double X_norm_inf;
	
	typedef double (*n1_p)[N];
	typedef double const (*n1c_p)[N];
	typedef double (*m1_p)[M];
	typedef double const (*m1c_p)[M];
	
	initializeUsingUserSpecification(1, N, 1, OtherBS, (n1_p) X);
	
	#pragma omp parallel
	#pragma omp single
	{
		#pragma omp task untied
		matrix_norm_inf(OtherBS, OtherBS, M, N, (mc_p) C, &C_norm_inf);
		#pragma omp task untied
		vector_norm_inf(OtherBS, N, X, &X_norm_inf);
	
		#pragma omp task untied
		matrix_vector_product_iterative(OtherBS, OtherBS, N, N, B, X, X2); // X2 := BX
		#pragma omp task untied
		matrix_vector_product_iterative(OtherBS, OtherBS, M, N, C, X, Z); // Z := CX
		
		#pragma omp taskwait
	}
	
	#pragma omp parallel
	#pragma omp single
	matrix_vector_product_iterative(OtherBS, OtherBS, M, N, A, X2, Y); // Y := A(X2)
	
	#pragma omp parallel
	#pragma omp single
	substract_iterative(1, M, 1, M, 1, M, // X3 := Y - Z === ABX - CX
		1, M,
		(m1_p) Y, (m1_p) Z, (m1_p) X3, OtherBS);
	
	double resid0;
	#pragma omp parallel
	#pragma omp single
	vector_norm_inf(OtherBS, M, X3, &resid0);
	
	free(X);
	free(X2);
	free(X3);
	free(Y);
	free(Z);
	
	double scaled_residual = resid0 / (C_norm_inf * X_norm_inf * N * DBL_EPSILON);
	
	if (scaled_residual >= 16.0) {
		fprintf(stderr, "Verification error: %g\n", scaled_residual);
		return 1;
	}
#endif
	
	if (getenv("MEASURE_TIME") == NULL) {
		// Performance in MFLOPS of a standard matrix multiplication
		printf ("%f", ((2.0 * (double) M * (double) N * (double) K)/GET_TIME()) );
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	return 0;
}

