#ifndef OMPSS_MEMORY_H
#define OMPSS_MEMORY_H


#ifndef _BSD_SOURCE
#define _BSD_SOURCE
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>


// FIXME: This should really be in a library. If memory is allocated in one file and freed in another this will fail.
typedef struct allocation_entry_struct {
	void *pointer;
	size_t size;
	struct allocation_entry_struct * next;
} allocation_entry_t;


static allocation_entry_t * __allocated_entries = NULL;


static void *aligned_malloc(size_t size) {
	size_t page_size = sysconf(_SC_PAGESIZE);
	size_t number_of_pages = (size + page_size - 1L) / page_size;
	size_t corrected_size = number_of_pages * page_size;
	
#if __MACH__
	char *new_memory = (char *)mmap(0, corrected_size * 2L, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON, -1, 0);
#else
	char *new_memory = (char *)mmap(0, corrected_size * 2L, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
#endif
	if (new_memory == MAP_FAILED) {
		perror("aligned_malloc");
		return NULL;
	}
	
	char *new_aligned_memory;
	size_t new_memory_unalignment = (size_t)new_memory % corrected_size;
	
	if (new_memory_unalignment == 0L) {
		new_aligned_memory = new_memory;
		munmap(new_memory + corrected_size, corrected_size);
	} else {
		new_aligned_memory = new_memory + corrected_size - new_memory_unalignment;
		int rc = munmap(new_memory, new_aligned_memory - new_memory);
		if (rc != 0) {
			perror("aligned_malloc::munmap1");
			return NULL;
		}
		rc = munmap(
			new_aligned_memory + corrected_size,
			new_memory + 2L * corrected_size - (new_aligned_memory + corrected_size));
		if (rc != 0) {
			perror("aligned_malloc::munmap2");
			return NULL;
		}
	}
	
	allocation_entry_t *newEntry = (allocation_entry_t *) malloc(sizeof(allocation_entry_t));
	newEntry->pointer = new_aligned_memory;
	newEntry->size = corrected_size;
	
	#pragma omp critical
	{
		newEntry->next = __allocated_entries;
		__allocated_entries = newEntry;
	}
	
	return new_aligned_memory;
}


static void aligned_free(void *pointer) {
	allocation_entry_t *entry;
	#pragma omp critical
	{
		// Find the entry
		allocation_entry_t **pentry = &__allocated_entries;
		while (*pentry != NULL) {
			entry = *pentry;
			
			if (entry->pointer == pointer) {
				break;
			}
			
			pentry = &(entry->next);
		}
		
		if (*pentry == NULL) {
			fprintf(stderr, "Invalid call to aligned_free with pointer %p\n", pointer);
			exit(1);
		}
		
		// Remove the entry
		*pentry = entry->next;
	}
	
	int rc = munmap(pointer, entry->size);
	free(entry);
	if (rc != 0) {
		perror("aligned_free::munmap");
	}
}


#if defined(OVERRIDE_MALLOC_AND_FREE)
#define malloc aligned_malloc
#define free aligned_free
#endif


#endif // OMPSS_MEMORY_H

