#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define TYPE double
#include "../../../utils/blas.h"
#include "../../../utils/blas_init.h"
#include "../../../utils/timer.h"
#include "regions/common/init.h"

#define OVERRIDE_MALLOC_AND_FREE
#include "memory.h"

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;


//
// Kernels
//

#pragma omp task inout(([N][N]A)[0;BS][0;BS]) priority(1) 
void dpotrf_tile(double *A, integer BS, integer N) {
	integer INFO;
	
	lapack_dpotrf("L", &BS, A, &N, &INFO);
}


#pragma omp task in(([N][N]A)[0;BS][0;BS], ([N][N]B)[0;BS][0;BS]) inout(([N][N]C)[0;BS][0;BS])
void dgemm_tile(double  *A, double *B, double *C, integer BS, integer N) {
	lapack_dgemm("N", "T", &BS, &BS, &BS, &dminusone, A, &N, B, &N, &done, C, &N);
}


#pragma omp task in(([N][N]T)[0;BS][0;BS]) inout(([N][N]B)[0;BS][0;BS])
void dtrsm_tile(double *T, double *B, integer BS, integer N) {
	lapack_dtrsm("R", "L", "T", "N", &BS, &BS, &done, T, &N, B, &N);
}


#pragma omp task in(([N][N]A)[0;BS][0;BS]) inout(([N][N]C)[0;BS][0;BS])
void dsyrk_tile( double *A, double *C, integer BS, integer N) {
	lapack_dsyrk("L", "N", &BS, &BS, &dminusone, A, &N, &done, C, &N);
}


void cholesky(integer N, integer BS, double A[N][N]) {
	for (integer j = 0; j < N; j+=BS) {
		for (integer k= 0; k< j; k+=BS) {
			for (integer i = j+BS; i < N; i+=BS) {
				// A[i,j] = A[i,j] - A[i,k] * (A[j,k])^t
				dgemm_tile( &A[k][i], &A[k][j], &A[j][i], BS, N);
			}
		}
		
		for (integer i = 0; i < j; i+=BS) {
			// A[j,j] = A[j,j] - A[j,i] * (A[j,i])^t
			dsyrk_tile( &A[i][j], &A[j][j], BS, N);
		}
		
		// Cholesky Factorization of A[j,j]
		dpotrf_tile( &A[j][j], BS, N);
		
		for (integer i = j+BS; i < N; i+=BS) {
			// A[i,j] <- A[i,j] = X * (A[j,j])^t
			dtrsm_tile( &A[j][j], &A[j][i], BS, N);
		}
	}
}


//--------------------------------------------------------------------------------


static integer N_;


static inline double initializeElement(long I, long J, long i, long j) {
	// Symetric and positive definite
	if ((I+i) == (J+j)) {
		return ((double) N_) + (I+J+i*3.0+j*5.0)/(10.0*N_);
	} else {
		return (I+J+i*3.0+j*5.0)/(10.0*N_);
	}
}


static void init(integer N, integer BS, double A[N][N]) {
	force_blas_init();
	initializeUsingUserSpecification(N, N, BS, BS, A);
}


int main(int argc, char *argv[]) {
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <N> <N> <block size> <block size> <index> <H|V|B|PW>\n", argv[0]);
		exit (1);
	}
	
	integer N = atol(argv[1]);
	integer BS = atol(argv[3]);
	parseInitializationType(argv[6]);
	
	N_ = N;
	
	double (*A)[N] = (double *) malloc (sizeof(double) * N * N);
	init(N, BS, A);
	#pragma omp taskwait
	
	START_TIME();
	cholesky(N, BS, A);
	#pragma omp taskwait
	STOP_TIME();
	
	if (getenv("MEASURE_TIME") == NULL) {
		printf("%f", (0.33*N*N*N+0.5*N*N+0.17*N)/GET_TIME() );
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	return 0;
}
