#ifndef USER_INITIALIZATION_BY_SHAPE_H
#define USER_INITIALIZATION_BY_SHAPE_H


#include <stdlib.h>
#include <string.h>


static enum {
	H_INIT,
	V_INIT,
	B_INIT,
	PW_INIT,
	SEQ_INIT
} initialization_type;


int initializationIsQuadratic() {
	return (initialization_type == B_INIT) || (initialization_type == PW_INIT);
}


#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
static inline TYPE initializeElement(long I, long J, long i, long j, long VN, long HN, long realVN, long realHN);
#else
static inline TYPE initializeElement(long I, long J, long i, long j);
#endif

#pragma omp task out(([realVN]data)[0;VBS][0;HBS])
static void zz_initializeBlock(long VN, long HN, long realVN, long realHN, long VBS, long HBS, long I, long J, TYPE data[realVN][realHN]) {
	for (long i=0; i < VBS; i++) {
		for (long j=0; j < HBS; j++) {
#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
			data[i][j] = initializeElement(I, J, i, j, VN, HN);
#else
			data[i][j] = initializeElement(I, J, i, j);
#endif
		}
	}
}


static void initialize(long VN, long HN, long realVN, long realHN, long VBS, long HBS, TYPE data[realVN][realHN]) {
	for (long I = 0; I < VN; I+=VBS) {
		for (long J = 0; J < HN; J+=HBS) {
			zz_initializeBlock(VN, HN, realVN, realHN, VBS, HBS, I, J, (TYPE (*)[realHN]) &data[I][J]);
		}
	}
}


#pragma omp task out(([realVN1]data1)[0;VBS1][0;HBS1], ([realVN2]data2)[0;VBS2][0;HBS2])
static void zz_initialize2Blocks(long VN1, long HN1, long realVN1, long realHN1, long VBS1, long HBS1, long I1, long J1, TYPE data1[realVN1][realHN1], long VN2, long HN2, long realVN2, long realHN2, long VBS2, long HBS2, long I2, long J2, TYPE data2[realVN2][realHN2]) {
	for (long i=0; i < VBS1; i++) {
		for (long j=0; j < HBS1; j++) {
#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
			data1[i][j] = initializeElement(I1, J1, i, j, VN1, HN1, realVN1, realHN1);
#else
			data1[i][j] = initializeElement(I1, J1, i, j);
#endif
		}
	}
	for (long i=0; i < VBS2; i++) {
		for (long j=0; j < HBS2; j++) {
#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
			data2[i][j] = initializeElement(I2, J2, i, j, VN2, HN2, realVN2, realHN2);
#else
			data2[i][j] = initializeElement(I2, J2, i, j);
#endif
		}
	}
}


#pragma omp task out(([realN]data1)[0;BS][0;BS], ([realN]data2)[0;BS][0;BS])
static void zz_initializeBlockPair(long N, long realN, long BS, long I, long J, TYPE data1[realN][realN], TYPE data2[realN][realN]) {
	for (long i=0; i < BS; i++) {
		for (long j=0; j < BS; j++) {
#ifdef PASS_DIMENSIONS_TO_INITIALIZATION_FUNCTION
			data1[i][j] = initializeElement(I, J, i, j, N, N, realN, realN);
			data2[j][i] = initializeElement(J, I, j, i, N, N, realN, realN);
#else
			data1[i][j] = initializeElement(I, J, i, j);
			data2[j][i] = initializeElement(J, I, j, i);
#endif
		}
	}
}


static void initializeByPairs(long N, long realN, long BS, TYPE data[realN][realN]) {
	for (long I = 0; I < N; I+=BS) {
		for (long J = 0; J < I; J+=BS) {
			zz_initializeBlockPair(N, realN, BS, I, J, (TYPE (*)[realN]) &data[I][J], (TYPE (*)[realN]) &data[J][I]);
		}
		zz_initializeBlock(N, N, realN, realN, BS, BS, I, I, (TYPE (*)[realN]) &data[I][I]);
	}
}


static void parseInitializationType(char const *initializazionTypeString) {
	if (strcmp(initializazionTypeString, "H") == 0) {
		initialization_type = H_INIT;
	} else if (strcmp(initializazionTypeString, "V") == 0) {
		initialization_type = V_INIT;
	} else if (strcmp(initializazionTypeString, "B") == 0) {
		initialization_type = B_INIT;
	} else if (strcmp(initializazionTypeString, "PW") == 0) {
		initialization_type = PW_INIT;
	} else if (strcmp(initializazionTypeString, "SEQ") == 0) {
		initialization_type = SEQ_INIT;
	} else {
		fprintf(stderr, "Error: unknown initialization type parameter.\n");
		exit(1);
	}
}


static void initializeUsingUserSpecification(long VN, long HN, long realVN, long realHN, long VBS, long HBS, TYPE data[realVN][realHN]) {
	switch (initialization_type) {
		case H_INIT:
			// Horizontal panels
			initialize(VN, HN, realVN, realHN, VBS, HN, data);
			break;
		case V_INIT:
			// Vertical panels
			initialize(VN, HN, realVN, realHN, VN, HBS, data);
			break;
		case B_INIT:
			// Blocks
			initialize(VN, HN, realVN, realHN, VBS, HBS, data);
			break;
		case PW_INIT:
			if (HN != VN) {
				fprintf(stderr, "Error: this problem does not allow pairwise initialization.\n");
				exit(1);
			}
			// Pairs of blocks around the diagonal
			initializeByPairs(HN, realHN, HBS, data);
			break;
		case SEQ_INIT:
			// Sequential
			initialize(VN, HN, realVN, realHN, VN, HN, data);
			break;
	}
}


#endif // USER_INITIALIZATION_BY_SHAPE_H
