#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdint.h>

#include <complex.h>
#ifdef I
#undef I
#endif

#include <fftw3.h>


#define TYPE double _Complex
#include "../../../utils/timer.h"
#include "regions/common/init.h"

#define OVERRIDE_MALLOC_AND_FREE
#include "memory.h"


#define WISDOM_FILE \
	(getenv("FFTW_WISDOM_FILE") != 0 ? getenv("FFTW_WISDOM_FILE") : "fftw_ompss.wisdom")


fftw_plan plan, inverse_plan;


double _Complex * Alloc_FFT_Matrix (long N) {
	double _Complex *ptr;
	
	ptr = malloc (N*sizeof(double _Complex));
	if (ptr == NULL) {
		fprintf (stderr, "Cannot obtain aligned memory buffer. Dying...\n");
		exit (0);
	}
	return ptr;
}


static inline double _Complex initializeElement(long I, long J, long i, long j) {
	return I+i+j;
}


//
// Tasks
//

#pragma omp task inout(([N_SQRT]data)[0;FFT_BS][0;N_SQRT]) priority(1)
static void FFT1D (long N_SQRT, long FFT_BS, double _Complex data[N_SQRT][N_SQRT]) {
	for (long i=0; i < FFT_BS; i++) {
		fftw_execute_dft(plan, &data[i][0], &data[i][0]);
	}
}


#pragma omp task inout(([N_SQRT]data)[0;FFT_BS][0;N_SQRT]) priority(1)
static void FFT1D_2 (long N_SQRT, long FFT_BS, double _Complex data[N_SQRT][N_SQRT]) {
	for (long i=0; i < FFT_BS; i++) {
		fftw_execute_dft(plan, &data[i][0], &data[i][0]);
	}
}


#pragma omp task inout(([N_SQRT]data)[0;TR_BS][0;TR_BS])
void trsp_blk(long N, long N_SQRT, long TR_BS, double _Complex data[N_SQRT][N_SQRT]);

#pragma omp task inout(([N_SQRT]data1)[0;TR_BS][0;TR_BS], ([N_SQRT]data2)[0;TR_BS][0;TR_BS])
void trsp_swap (long N, long N_SQRT, long TR_BS, double _Complex data1[N_SQRT][N_SQRT], double _Complex data2[N_SQRT][N_SQRT]);

#pragma omp task inout(([N_SQRT]data)[0;TR_BS][0;TR_BS])
void tw_trsp_blk(long N, long N_SQRT, long TR_BS, long I, double _Complex data[N_SQRT][N_SQRT]);

#pragma omp task inout(([N_SQRT]data1)[0;TR_BS][0;TR_BS], ([N_SQRT]data2)[0;TR_BS][0;TR_BS])
void tw_trsp_swap (long N, long N_SQRT, long TR_BS, long I, long J, double _Complex data1[N_SQRT][N_SQRT], double _Complex data2[N_SQRT][N_SQRT]);


void FFT_1D (long N, long N_SQRT, long FFT_BS, long TR_BS, double _Complex A[N_SQRT][N_SQRT]) {
	typedef double _Complex (*m_p)[N_SQRT];
	
	// Transpose
	for (long I=0; I<N_SQRT; I+=TR_BS) {
		trsp_blk (N, N_SQRT, TR_BS, (m_p) &A[I][I]);
		for (long J=I+TR_BS; J<N_SQRT; J+=TR_BS)
			trsp_swap (N, N_SQRT, TR_BS, (m_p) &A[I][J], (m_p) &A[J][I]);
	}
	
	// First FFT round
	for (long J=0; J<N_SQRT; J+=FFT_BS)
		FFT1D(N_SQRT, FFT_BS, (m_p) &A[J][0]);
	
	// Twiddle and Transpose
	for (long I=0; I<N_SQRT; I+=TR_BS) {
		tw_trsp_blk (N, N_SQRT, TR_BS, I, (m_p) &A[I][I]);
		for (long J=I+TR_BS; J<N_SQRT; J+=TR_BS)
			tw_trsp_swap (N, N_SQRT, TR_BS, I, J, (m_p) &A[I][J], (m_p) &A[J][I]);
	}
	
	// Second FFT round
	for (long J=0; J<N_SQRT; J+=FFT_BS)
		FFT1D_2(N_SQRT, FFT_BS, (m_p) &A[J][0]);
	
	// Transpose
	for (long I=0; I<N_SQRT; I+=TR_BS) {
		trsp_blk (N, N_SQRT, TR_BS, (m_p) &A[I][I]);
		for (long J=I+TR_BS; J<N_SQRT; J+=TR_BS)
			trsp_swap (N, N_SQRT, TR_BS, (m_p) &A[I][J], (m_p) &A[J][I]);
	}
}


static void setup_fftw_plan(long N_SQRT) {
	FILE *file;
	
	file = fopen(WISDOM_FILE, "r");
	if (file != NULL) {
		fftw_import_wisdom_from_file(file);
		fclose(file);
	}
	
	// Prepare a plan for a row
	double _Complex *data = fftw_malloc(sizeof(double _Complex) * N_SQRT);
	
	// Train the FFTW engine
	plan = fftw_plan_dft_1d(N_SQRT, data, data, FFTW_FORWARD, FFTW_PATIENT);
	
	fftw_free(data);
	
	// Save the wisdom
	file = fopen(WISDOM_FILE, "w");
	fftw_export_wisdom_to_file(file);
	fclose(file);
}


static void setup_fftw_inverse_plan(long N) {
	FILE *file;
	
	file = fopen(WISDOM_FILE, "r");
	if (file != NULL) {
		fftw_import_wisdom_from_file(file);
		fclose(file);
	}
	
	// Prepare a plan for a row
	double _Complex *data = fftw_malloc(sizeof(double _Complex) * N);
	
	// Train the FFTW engine
	inverse_plan = fftw_plan_dft_1d(N, data, data, FFTW_FORWARD, FFTW_ESTIMATE);
	
	fftw_free(data);
	
	// Save the wisdom
	file = fopen(WISDOM_FILE, "w");
	fftw_export_wisdom_to_file(file);
	fclose(file);
}


int main(int argc, char **argv)
{
	double _Complex *A __attribute__ ((aligned (128)));
	
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <N> <N> <FFT block size> <transpose block size> <index> <H|V|B|PW|SEQ>\n", argv[0]);
		return 1;
	}
	
	long N_SQRT = atol(argv[1]);
	long N = N_SQRT*N_SQRT;
	long FFT_BS = atol(argv[3]);
	long TR_BS = atol(argv[4]);
	parseInitializationType(argv[6]);
	
	if (FFT_BS > N_SQRT) {
		fprintf(stderr, "Error: either the FFT block size must be <= %li or elements >= %li\n", N_SQRT, FFT_BS*FFT_BS);
		return 1;
	}
	if (TR_BS > N_SQRT) {
		fprintf(stderr, "Error: either the transpose block size must be <= %li or elements >= %li\n", N_SQRT, TR_BS*TR_BS);
		return 1;
	}
	
	A = Alloc_FFT_Matrix(N);
#if CHECK
	setup_fftw_inverse_plan(N);
#endif
	
	initializeUsingUserSpecification(N_SQRT, N_SQRT, FFT_BS, FFT_BS, (void *) A);
	setup_fftw_plan(N_SQRT);
	
	#pragma omp taskwait
	START_TIME();
	FFT_1D (N, N_SQRT, FFT_BS, TR_BS, (void *)A);
	#pragma omp taskwait
	STOP_TIME();
	
	if (getenv("MEASURE_TIME") == NULL) {
		printf("%f", (5.0*(double)N*log(N)/log(2.0))/GET_TIME());
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
#ifdef CHECK
	fftw_execute_dft(inverse_plan, A, A);
#endif
	
	
	free(A);
}

