#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../../../utils/timer.h"

#define OVERRIDE_MALLOC_AND_FREE
#include "memory.h"


// N and MIN_BS must be powers of 2
long N;
long MIN_SORT_SIZE;
long MIN_MERGE_SIZE;

#define T int


#pragma omp task inout([n]data)
void basicsort(long n, T *data);


#pragma omp task in([n]left, [n]right) out(([n*2]result)[start;length])
void merge_task(long n, T *left, T *right, T *result, long start, long length);


void merge_rec(long n, T left[n], T right[n], T result[n*2], long start, long length) {
	if (length <= MIN_MERGE_SIZE) {
		// Base case
		merge_task(n, left, right, result, start, length);
	} else {
		merge_rec(n, left, right, result, start, length/2);
		merge_rec(n, left, right, result, start + length/2, length/2);
	}
}


void merge(long n, T left[n], T right[n], T result[n*2]) {
	long length = n * 2L;
	if (length <= MIN_MERGE_SIZE) {
		// Base case
		merge_task(n, left, right, result, 0, n);
	} else {
		merge_rec(n, left, right, result, 0, length/2L);
		merge_rec(n, left, right, result, length/2L, length/2L);
	}
}


void multisort(long n, T data[n], T tmp[n]) {
	if (n >= MIN_SORT_SIZE*4L) {
		// Recursive decomposition
		multisort(n/4L, &data[0], &tmp[0]);
		multisort(n/4L, &data[n/4L], &tmp[n/4L]);
		multisort(n/4L, &data[n/2L], &tmp[n/2L]);
		multisort(n/4L, &data[3L*n/4L], &tmp[3L*n/4L]);
		
		merge_rec(n/4L, &data[0], &data[n/4L], &tmp[0], 0, n/2L);
		merge_rec(n/4L, &data[n/2L], &data[3L*n/4L], &tmp[n/2L], 0, n/2L);
		
		merge_rec(n/2L, &tmp[0], &tmp[n/2L], &data[0], 0, n);
	} else {
		// base case
		basicsort(n, data);
	}
}


#pragma omp task out(([rows]data)[0;rows][0;columns], ([rows]tmp)[rows][0;columns])
static void zz_init_and_touch(long rows, long row_length, long columns, T data[rows][row_length], T tmp[rows][row_length]) {
	T last_rand = rand();
	
	for (long row = 0; row < rows; row++) {
		for (long column = 0; column < columns; column++) {
			data[row][column] = last_rand;
			last_rand = (last_rand * (row + 1L) * (column + 1L) * 104723L) % N;
			tmp[row][column] = 0; 
		}
	}
}


static void init_and_touch(long N, long BS, long NB, T *data, T *tmp) {
	for (long i=0; i < NB; i++) {
		zz_init_and_touch(N/(BS*NB), BS*NB, BS, &data[i*BS], &tmp[i*BS]);
	}
}


int main(int argc, char **argv) {
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <N/2^20> <N/2^20> <sort block size/2^20> <merge block size/2^20> <threads> <M|S|M2|S2|T>\n", argv[0]);
		return 1;
	}
	
	N = atol(argv[1]) * (1024L * 1024L);
	MIN_SORT_SIZE = atol(argv[3]) * (1024L * 1024L);
	MIN_MERGE_SIZE = atol(argv[4]) * (1024L * 1024L);
	long threads = atol(argv[5]);
	
	enum {
		MERGE_SIZE_INIT,
		SORT_SIZE_INIT,
		MERGE_SIZE2_INIT,
		SORT_SIZE2_INIT,
		THREAD_COUNT_INIT,
		MEMORY_NODE_INIT
	} initialization_type;
	
	if (strcmp(argv[6], "M") == 0) {
		initialization_type = MERGE_SIZE_INIT;
	} else if (strcmp(argv[6], "S") == 0) {
		initialization_type = SORT_SIZE_INIT;
	} else if (strcmp(argv[6], "M2") == 0) {
		initialization_type = MERGE_SIZE2_INIT;
	} else if (strcmp(argv[6], "S2") == 0) {
		initialization_type = SORT_SIZE2_INIT;
	} else if (strcmp(argv[6], "T") == 0) {
		initialization_type = THREAD_COUNT_INIT;
	} else if (strcmp(argv[6], "MN") == 0) {
		initialization_type = MEMORY_NODE_INIT;
	} else {
		fprintf(stderr, "Error: unknown initialization type parameter.\n");
		exit(1);
	}
	
	// Data
	T *data = malloc(N*sizeof(T));
	T *tmp = malloc(N*sizeof(T));
	
	switch (initialization_type) {
		case MERGE_SIZE_INIT:
			init_and_touch(N, MIN_MERGE_SIZE, threads, data, tmp);
			break;
		case SORT_SIZE_INIT:
			init_and_touch(N, MIN_SORT_SIZE, threads, data, tmp);
			break;
		case MERGE_SIZE2_INIT:
			init_and_touch(N, MIN_MERGE_SIZE*2L, threads, data, tmp);
			break;
		case SORT_SIZE2_INIT:
			init_and_touch(N, MIN_SORT_SIZE*2L, threads, data, tmp);
			break;
		case THREAD_COUNT_INIT:
			init_and_touch(N, N/threads, threads, data, tmp);
			break;
		case MEMORY_NODE_INIT:
			{
				long memoryNodes = (threads + 4 - 1) / 4;
				fprintf(stderr, "Warning: Assuming 4 cores per NUMA node\n");
				init_and_touch(N, N/memoryNodes, memoryNodes, data, tmp);
			}
			break;
	}
	#pragma omp taskwait
	
	START_TIME();
	multisort(N, data, tmp);
	#pragma omp taskwait
	STOP_TIME();
	
	if (getenv("MEASURE_TIME") == NULL) {
		// Melements/s
		printf("%f", ((double)N) / ((double)GET_TIME()));
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	free(data);
	free(tmp);
	
	return 0;
}
