#!/bin/sh

if [ "${V}" = "1" ] ; then
	set -x
fi


usage() {
	cat << EOF
Usage: $0 [NP] [S|M|L] [directory] [executable]
	NP:	Number of threads/processes
	S:	Run the benchmark with a small size
	M:	Run the benchmark with a medium size
	L:	Run the benchmark with a large size
	
	--help:	This help message
	
	MPIRUN_PARAMS can be set to any extra parameters that should be passed to mpirun.
	NX_INSTRUMENTATION can be set to "extrae" to execute the benchmarks with tracing.
	
	QUIET deactivates all output if not empty
	V makes the output verbose if not empty
EOF
}

basedir=$(dirname $0)
scriptdir=${basedir}/../scripts

if [ "$#" != "4" ] ; then
	usage
	exit 1
fi

export NP=$1
problem_size=$2
bench=$3
binary=$4

if [ "${NP}" = "--help" ] ; then
	usage
	exit
fi

# Verify the problem size
case "${problem_size}" in
S|M|L)
	;;
*)
	echo Error: invalid parameter \"${param}\". 1>&2
	usage
	exit 1
	;;
esac


generate_extrae_config() {
	local executable_path="$1"
	shift
	local executable=$(basename "${executable_path}")
	
	local DO_OPENMP=no
	local DO_PTHREAD=no
	local DO_MPI=no
	
	if [ -z "${EXTRAE_HOME}" ] ; then
		export EXTRAE_HOME=$(dirname $(dirname $(which extrae)))
		
		if [ -z "${EXTRAE_HOME}" ] ; then
			echo "Error: either set EXTRAE_HOME, or add extrae to your path." 1>&2
			exit 1
		fi
	fi
	
	local trace_path="traces/"$(dirname "${executable_path}")
	mkdir -p "${trace_path}"
	local trace_real_path=$(readlink -f "${trace_path}")
	
	local marshalled_parameters=$(echo "$*" | sed 's/ /-/g')
	case "${executable}" in
	*ompss*)
		;;
	*omp*)
		DO_OPENMP=yes
		;;
	*threaded_lapack*)
		DO_PTHREAD=yes
		;;
	esac
	
	case "${executable}" in
	*mpi*)
		DO_MPI=yes
		;;
	esac
	
cat > .extrae.xml << EOF
<?xml version='1.0'?>

<trace enabled="yes"
 home="${EXTRAE_HOME}"
 initial-mode="user"
 type="paraver"
 xml-parser-id="Id: xml-parse.c 2014 2013-07-30 08:13:03Z harald $"
>

  <openmp enabled="${DO_OPENMP}">
    <locks enabled="no" />
    <counters enabled="yes" />
  </openmp>

  <pthread enabled="${DO_PTHREAD}">
    <locks enabled="no" />
    <counters enabled="yes" />
  </pthread>

  <callers enabled="yes">
    <mpi enabled="${DO_MPI}">1-3</mpi>
    <sampling enabled="no">1-5</sampling>
  </callers>

  <user-functions enabled="no" list="/home/bsc41/bsc41273/user-functions.dat" exclude-automatic-functions="no">
    <counters enabled="yes" />
  </user-functions>

  <counters enabled="yes">
    <cpu enabled="yes" starting-set-distribution="1">
      <set enabled="yes" domain="all" changeat-globalops="5">
        PAPI_TOT_INS,PAPI_TOT_CYC,PAPI_L1_DCM
        <sampling enabled="no" frequency="100000000">PAPI_TOT_CYC</sampling>
      </set>
      <set enabled="yes" domain="user" changeat-globalops="5">
        PAPI_TOT_INS,PAPI_FP_INS,PAPI_TOT_CYC
      </set>
    </cpu>

    <network enabled="no" />

    <resource-usage enabled="no" />

    <memory-usage enabled="no" />
  </counters>

  <storage enabled="yes">
    <trace-prefix enabled="no">${executable}-${NP}threads-${marshalled_parameters}</trace-prefix>
    <size enabled="no">5</size>
    <temporal-directory enabled="no">/scratch</temporal-directory>
    <final-directory enabled="no">${trace_real_path}</final-directory>
    <gather-mpits enabled="no" />
  </storage>

  <buffer enabled="yes">
    <size enabled="yes">500000</size>
    <circular enabled="no" />
  </buffer>

  <trace-control enabled="no">
    <file enabled="no" frequency="5M">/gpfs/scratch/bsc41/bsc41273/control</file>
    <global-ops enabled="no"></global-ops>
    <remote-control enabled="no">
      <signal enabled="no" which="USR1"/>
    </remote-control>
  </trace-control>

  <others enabled="no">
    <minimum-time enabled="no">10M</minimum-time>
  </others>

  <bursts enabled="no">
    <threshold enabled="yes">500u</threshold>
    <mpi-statistics enabled="yes" />
  </bursts>

  <cell enabled="no">
    <spu-file-size enabled="yes">5</spu-file-size>
    <spu-buffer-size enabled="yes">64</spu-buffer-size>
    <spu-dma-channel enabled="no">2</spu-dma-channel>
  </cell>

  <sampling enabled="no" type="default" period="50m" variability="10m" />

  <dynamic-memory enabled="no" />

  <input-output enabled="no" />

  <merge enabled="yes" 
    synchronization="default"
    tree-fan-out="16"
    max-memory="512"
    joint-states="yes"
    keep-mpits="no"
    sort-addresses="yes"
    overwrite="no"
    binary="${executable_path}"
  >${trace_real_path}/${executable}-${NP}threads-${marshalled_parameters}.prv</merge>

</trace>
EOF
	
if [ -z "${QUIET}" -a ! -z "${V}" ] ; then
	echo ===== EXTRAE .extrae.xml BEGIN ===== 1>&2
	cat .extrae.xml
	echo ===== EXTRAE .extrae.xml END ===== 1>&2
fi
	echo "${trace_path}/${executable}-${NP}threads-${marshalled_parameters}"
}


clear_interfering_thread_envvars() {
	export OPENBLAS_NUM_THREADS=1
	export GOTO_NUM_THREADS=1
}


basedir=$(dirname $0)
src_dir_entry="${basedir}/${bench}"
binary=$(echo ${bench} | awk -F / '{ res=""; for (i=NF; i>1; i--) res = res $i "_"; print res $1; }')
build_dir_entry="${bench}"
executable="${build_dir_entry}/${binary}"
algorithm=$(basename "${bench}")

# Implementation-specific parameters
sizer_script="${src_dir_entry}/get_parameters_for_size"
if [ ! -x "${sizer_script}" ] ; then
	# Or parameters common to all the implementations of the benchmark
	sizer_script="${basedir}/common/${algorithm}/get_parameters_for_size"
fi

if [ ! -x "${executable}" ] ; then
	if [ -z "${SKIP_NOT_READY}" ] ; then
		echo "Error: could not find the '${executable}' executable" 1>&2
		exit 1
	else
		exit 0
	fi
fi

if [ ! -x "${sizer_script}" ] ; then
	if [ -z "${SKIP_NOT_READY}" ] ; then
		echo "Error: could not find the script to determine the problem size" 1>&2
		exit 1
	else
		exit 0
	fi
fi

if [ ! -z "${USE_FIXED_SIZE}" ] ; then
	# In this case we need a fixed problem size for any combination of threads
	bench_parameters=$(env NP=8 "${sizer_script}" ${problem_size} | awk '{print $1 " " $2 " " $3 " " $4 " '${NP}' " $6;}')
else
	# When testing we do not want to spend too much time with few or only one thread
	# so we decide the problem size in terms of the number of threads
	bench_parameters=$("${sizer_script}" ${problem_size})
fi

if [ "${NX_INSTRUMENTATION}" = "extrae" ] ; then
	export EXTRAE_CONFIG_FILE=.extrae.xml
	if [ -z "${QUIET}" ] ; then
		echo "Executing (with instrumentation): ${executable} (${problem_size})	${bench_parameters}" 1>&2
		echo -n "	Output trace prefix: " 1>&2
	fi
	generate_extrae_config "${executable}" ${bench_parameters}
else
	if [ -z "${QUIET}" ] ; then
		echo Executing: "${executable} (${problem_size})	${bench_parameters}" 1>&2
	fi
fi

case "${binary}" in
*threaded_lapack*)
	export OMP_NUM_THREADS="${NP}"
	;;
*ompss*)
	export NX_ARGS="${NX_ARGS} --smp-workers=${NP}"
	uses_ompss=yes
	clear_interfering_thread_envvars
	;;
*omp*)
	export OMP_NUM_THREADS="${NP}"
	clear_interfering_thread_envvars
	;;
*)
	clear_interfering_thread_envvars
	;;
esac

case "${binary}" in
*regions*)
	export NX_DEPS=regions
	;;
esac

case "${binary}" in
*threaded_fftw*)
	export FFTW_WISDOM_FILE="${HOME}/.threaded_fftw.wisdom"
	;;
*mpi_fftw*)
	export FFTW_WISDOM_FILE="${HOME}/.mpi_fftw.wisdom"
	;;
*fftw*)
	export FFTW_WISDOM_FILE="${HOME}/.fftw.wisdom"
	;;
esac

if [ "${NX_INSTRUMENTATION}" = "extrae" -a "${uses_ompss}" != "yes" ] ; then
	if ldd "${executable}" | grep -q omp ; then
		uses_openmp=yes
	fi
	if ldd "${executable}" | grep -q mpi ; then
		uses_mpi=yes
	fi
	if ldd "${executable}" | grep -q pthread ; then
		uses_pthreads=yes
	fi
	if [ "${uses_openmp}" = "yes" ] ; then
		if [ "${uses_mpi}" = "yes" ] ; then
			TRACER_LIB_DISCRIMINANT=ompi
		else
			TRACER_LIB_DISCRIMINANT=omp
		fi
	elif [ "${uses_mpi}" = "yes" ] ; then
		if [ "${uses_pthreads}" = "yes" ] ; then
			TRACER_LIB_DISCRIMINANT=ptmpi
		else
			TRACER_LIB_DISCRIMINANT=mpi
		fi
	elif [ "${uses_pthreads}" = "yes" ] ; then
		TRACER_LIB_DISCRIMINANT=pt
	else
		TRACER_LIB_DISCRIMINANT=seq
	fi
	
	unset uses_openmp uses_mpi uses_pthreads
	launcher="env LD_PRELOAD=${EXTRAE_HOME}/lib/lib${TRACER_LIB_DISCRIMINANT}trace.so"
fi

if [ ! -z "${QUIET}" ] ; then
	MPIRUN_PARAMS="${MPIRUN_PARAMS} -q"
fi

case "${binary}" in
*mpi*)
	result=$(mpirun -np ${NP} ${MPIRUN_PARAMS} ${launcher} "${executable}" ${bench_parameters})
	rc=$?
	;;
*)
	result=$(${launcher} "${PWD}/${executable}" ${bench_parameters})
	rc=$?
	;;
esac

if [ "${result}" = "0" -o -z "${result}" ] ; then
	rc=2
fi

rm -Rf set-0 .extrae.xml XMLFile* TRACE.*

if [ "$rc" = "0" ] ; then
	echo -n "${result}"
fi

exit $rc
