#ifndef USER_INITIALIZATION_BY_SHAPE_H
#define USER_INITIALIZATION_BY_SHAPE_H


#include <stdlib.h>
#include <string.h>


static enum {
	H_INIT,
	V_INIT,
	B_INIT,
	PW_INIT,
	SEQ_INIT
} initialization_type;


static inline TYPE initializeElement(long I, long J, long i, long j);


static void zz_initializeBlock(long VN, long HN, long VBS, long HBS, long I, long J, TYPE data[VN][HN]) {
	for (long i=0; i < VBS; i++) {
		for (long j=0; j < HBS; j++) {
			data[i][j] = initializeElement(I, J, i, j);
		}
	}
}


static void zz_initializeBlockPair(long N, long BS, long I, long J, TYPE data1[N][N], TYPE data2[N][N]) {
	for (long i=0; i < BS; i++) {
		for (long j=0; j < BS; j++) {
			data1[i][j] = initializeElement(I, J, i, j);
			data2[j][i] = initializeElement(J, I, j, i);
		}
	}
}


#if USE_SILLYMP
#include "sillymp.h"

typedef struct {
	long VN, HN, VBS, HBS, I, J;
	TYPE *data;
} zz_initializeBlock_arg_t;

void zz_initializeBlockWorkWrapper(void *opaque_arg) {
	zz_initializeBlock_arg_t *arg = opaque_arg;
	
	zz_initializeBlock(arg->VN, arg->HN, arg->VBS, arg->HBS, arg->I, arg->J, arg->data);
	free(arg);
}

void zz_initializeBlockWrapper(long VN, long HN, long VBS, long HBS, long I, long J, TYPE data[VN][HN]) {
	zz_initializeBlock_arg_t *arg = malloc(sizeof(zz_initializeBlock_arg_t));
	
	arg->VN = VN;
	arg->HN = HN;
	arg->VBS = VBS;
	arg->HBS = HBS;
	arg->I = I;
	arg->J = J;
	arg->data = data;
	
	sillymp_spawn(zz_initializeBlockWorkWrapper, arg);
}


typedef struct {
	long N, BS, I, J;
	TYPE *data1;
	TYPE *data2;
} zz_initializeBlockPair_arg_t;

void zz_initializeBlockPairWorkWrapper(void *opaque_arg) {
	zz_initializeBlockPair_arg_t *arg = opaque_arg;
	
	zz_initializeBlockPair(arg->N, arg->BS, arg->I, arg->J, arg->data1, arg->data2);
	free(arg);
}

void zz_initializeBlockPairWrapper(long N, long BS, long I, long J, TYPE data1[N][N], TYPE data2[N][N]) {
	zz_initializeBlockPair_arg_t *arg = malloc(sizeof(zz_initializeBlockPair_arg_t));
	
	arg->N = N;
	arg->BS = BS;
	arg->I = I;
	arg->J = J;
	arg->data1 = data1;
	arg->data2 = data2;
	
	sillymp_spawn(zz_initializeBlockPairWorkWrapper, arg);
}

#define zz_initializeBlock zz_initializeBlockWrapper
#define zz_initializeBlockPair zz_initializeBlockPairWrapper
#endif


static void initialize(long VN, long HN, long VBS, long HBS, TYPE data[VN][HN]) {
	for (long I = 0; I < VN; I+=VBS) {
		for (long J = 0; J < HN; J+=HBS) {
			zz_initializeBlock(VN, HN, VBS, HBS, I, J, (TYPE (*)[HN]) &data[I][J]);
		}
	}
}


static void initializeByPairs(long N, long BS, TYPE data[N][N]) {
	for (long I = 0; I < N; I+=BS) {
		for (long J = 0; J < I; J+=BS) {
			zz_initializeBlockPair(N, BS, I, J, (TYPE (*)[N]) &data[I][J], (TYPE (*)[N]) &data[J][I]);
		}
		zz_initializeBlock(N, N, BS, BS, I, I, (TYPE (*)[N]) &data[I][I]);
	}
}


static void parseInitializationType(char const *initializazionTypeString) {
	if (strcmp(initializazionTypeString, "I") == 0) {
#if USE_SILLYMP
		// Interleaved memory allocation
		initialization_type = H_INIT;
#else
		fprintf(stderr, "Error: Interleaving not available without sillymp.\n");
		exit(1);
#endif
	} else if (strcmp(initializazionTypeString, "H") == 0) {
#if USE_SILLYMP
		sillymp_disable_memory_interleaving();
#endif
		initialization_type = H_INIT;
	} else if (strcmp(initializazionTypeString, "V") == 0) {
#if USE_SILLYMP
		sillymp_disable_memory_interleaving();
#endif
		initialization_type = V_INIT;
	} else if (strcmp(initializazionTypeString, "B") == 0) {
#if USE_SILLYMP
		sillymp_disable_memory_interleaving();
#endif
		initialization_type = B_INIT;
	} else if (strcmp(initializazionTypeString, "PW") == 0) {
#if USE_SILLYMP
		sillymp_disable_memory_interleaving();
#endif
		initialization_type = PW_INIT;
	} else if (strcmp(initializazionTypeString, "SEQ") == 0) {
#if USE_SILLYMP
		sillymp_disable_memory_interleaving();
#endif
		initialization_type = SEQ_INIT;
	} else {
		fprintf(stderr, "Error: unknown initialization type parameter.\n");
		exit(1);
	}
}


static void initializeUsingUserSpecification(long VN, long HN, long VBS, long HBS, TYPE data[VN][HN]) {
	switch (initialization_type) {
		case H_INIT:
			// Horizontal panels
			initialize(VN, HN, VBS, HN, data);
			break;
		case V_INIT:
			// Vertical panels
			initialize(VN, HN, VN, HBS, data);
			break;
		case B_INIT:
			// Blocks
			initialize(VN, HN, VBS, HBS, data);
			break;
		case PW_INIT:
			if (VN != HN) {
				fprintf(stderr, "Error: this problem does not allow pairwise initialization.\n");
				exit(1);
			}
			// Pairs of blocks around the diagonal
			initializeByPairs(VN, VBS, data);
			break;
		case SEQ_INIT:
			// Sequential
			initialize(VN, HN, VN, HN, data);
			break;
	}
}


#endif // USER_INITIALIZATION_BY_SHAPE_H
