#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define TYPE double
#include "../../utils/blas.h"
#include "../../utils/blas_init.h"
#include "../../utils/timer.h"

#if _OPENMP
#include "common/flat_init.h"
#else
#include "common/flat_init.h"
#endif
#include "../../utils/timer.h"

#if USE_SILLYMP
#include "sillymp.h"
#endif

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;


//
// Tasks of the main algorithm
//

static void dgetrf_tile(integer M, integer N, integer LDA, double A[N][LDA], integer IPIV[min(M, N)]) {
	integer INFO;
	lapack_dgetrf(&M, &N, (double *) A, &LDA, IPIV, &INFO);
}

void dlaswp_tile(integer max_rows, integer N, integer LDA, double A[N][LDA], integer K1, integer K2, integer const IPIV[K2+1]) {
	K1++; K2++;
	lapack_dlaswp(&N, (double *) A, &LDA, &K1, &K2, IPIV, &ONE);
}


static void dtrsv_L_tile(integer NB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	lapack_dtrsv("L", "N", (check_diagonal ? "N" : "U"), &NB, (const double *) A, &N, X, &ONE);
}

static void dtrsv_U_tile(integer NB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	lapack_dtrsv("U", "N", (check_diagonal ? "N" : "U"), &NB, (double *) A, &N, X, &ONE);
}


static void pivot_vector(integer N, double v[N], integer const IPIV[N]) {
	for (int i=0; i < N; i++) {
		double tmp = v[i];
		v[i] = v[IPIV[i]-1];
		v[IPIV[i]-1] = tmp;
	}
}




void tiled_dgetrf(integer NB, integer MINNB, integer M, integer N, integer LDA, double A[N][LDA], integer IPIV[], int RL) {
	typedef double (*a_p)[LDA];
	typedef const double (*ac_p)[LDA];
	
	dgetrf_tile(M, N, LDA, A, IPIV);
}


//
// Tasks of the verification
//

static void dgemv_tile(integer M, integer N, double alpha, integer LDA, double const A[N][LDA], double const X[N], double beta, double Y[M]) {
	lapack_dgemv("N", &M, &N, &alpha, (const double *) A, &LDA, X, &ONE, &beta, Y, &ONE);
}


static void idamax_task(integer N, const double X[N], double *result) {
	integer local_max = lapack_idamax(&N, X, &ONE);
	
	*result = max(*result, fabs(X[local_max-1]));
}


static void dasum_tile(integer N, integer LDA, const double X[N][LDA], double result[N]) {
	double local_result[N];
	
	for (integer j=0; j < N; j++)
		local_result[j] = 0.0;
	
	for (integer i=0; i < N; i++)
		for (integer j=0; j < N; j++)
			local_result[j] += fabs(X[i][j]);
	
	for (integer j=0; j < N; j++) {
		result[j] += local_result[j];
	}
}


//
// Decomposer functions for the verification
//

static void smpss_dtrsv_U(integer NB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	dtrsv_U_tile(N, N, A, X, check_diagonal);
}


static void smpss_dtrsv_L(integer NB, integer N, double const A[N][N], double X[N], _Bool check_diagonal) {
	dtrsv_L_tile(N, N, A, X, check_diagonal);
}


static void vector_norm_inf(integer NB, integer N, double const V[N], double *result) {
	*result = 0.0;
	
	idamax_task(N, V, result);
}


static void matrix_norm_inf(integer NB, integer M, integer N, double const A[N][M], double *result) {
	typedef const double (*ac_p)[N];
	
	double *sums = malloc( ((long)M) * sizeof(double));
	
	for (long j=0; j < M; j++)
		sums[j] = 0.0;
	
	dasum_tile(M, N, A, sums);
	
	*result = 0.0;
	idamax_task(M, sums, result);
	
	free(sums);
}


//
// Main code
//


static inline double initializeElement(long I, long J, long i, long j) {
	static __thread uint64_t seed = 0;
	
	if ((i == 0) && (j == 0)) {
		seed = (I+i+1) * (J+j+1);
	}
	
	double result = (~0UL)/2UL;
	result -= seed;
	result /= ((~0UL / 2UL) + 1UL);
	seed = 6364136223846793005UL * seed + 1442695040888963407UL;
	
	return result;
}


// NOTE: This function is called a sencond time for verification and must generate exactly the same sequence of random numbers
static void initialize_data(integer NB, integer N, double A[N][N], double B[N]) {
#if USE_SILLYMP
	sillymp_init(atoi(getenv("OMP_NUM_THREADS")));
#endif
	initializeUsingUserSpecification(N, N, NB, NB, A);
#if USE_SILLYMP
	sillymp_end();
#endif
	
	srand(1);
	for (int i=0; i < N; i++)
		B[i] = ((double)RAND_MAX - 2.0*(double)rand()) / (double)RAND_MAX;
}


static void dump_matrix(integer M, integer N, double A[M][N]) {
	for (integer i=0; i < M ; i++) {
		for (integer j=0; j < N; j++) {
			printf("%g\t", A[i][j]);
		}
		printf("\n");
	}
}


static void dump_int_matrix(integer M, integer N, integer A[M][N]) {
	for (integer i=0; i < M ; i++) {
		for (integer j=0; j < N; j++) {
			printf("%ld\t", (integer)A[i][j]);
		}
		printf("\n");
	}
}


int main(int argc, char* argv[]) {
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <N> <N> <BS> <BS> <index> <I|H|V|B|PW|SEQ>\n", argv[0]);
		return 1;
	}
	
	integer N = atol(argv[1]);
	integer NB = atol(argv[3]);
	integer MINNB = atol(argv[4]);
	parseInitializationType(argv[6]);
	
	typedef double (*a_p)[N];
	typedef const double (*ac_p)[N];
	
	integer *IPIV = malloc(N*sizeof(integer)*2); // GotoBlas2 reads 2 elements after N
	a_p A = malloc(((long)N)*((long)N)*sizeof(double));
	double *B = malloc(((long)N)*sizeof(double));
	
	initialize_data(NB, N, A, B);
	START_TIME();
	
	// 1. LU factorization
	tiled_dgetrf(NB, MINNB, N, N, N, A, IPIV, 2);
	
	// 2. Solve the system with the LU factorization
	// 2.1 Pivot vector B
	pivot_vector(N, B, IPIV);
	// 2.2 Forward elimination
	smpss_dtrsv_L(max(1, NB/4), N, (ac_p) A, B, 0);
	// 2.3 Backwards substitution
	smpss_dtrsv_U(max(1, NB/4), N, (ac_p) A, B, 1);
	
	STOP_TIME();
	
	if (getenv("MEASURE_TIME") == NULL) {
		printf ("%f", (2.0*N*N*N/3.0+3.0*N*N/2.0)/GET_TIME());
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	double *X = B;
	B = malloc(((long)N)*sizeof(double));
	initialize_data(NB, N, A, B);
	
	double *v = malloc(((long)N)*sizeof(double));
	
	// v := B
	memcpy(v, B, ((long)N)*sizeof(double));
	
	// v := Ax - v
	lapack_dgemv("N", &N, &N, &done, (const double *)A, &N, X, &ONE, &dminusone, v, &ONE);
	
	double resid0;
	double A_norm_inf;
	double X_norm_inf;
	double B_norm_inf;
	
	vector_norm_inf(NB, N, v, &resid0);
	
	matrix_norm_inf(NB, N, N, (ac_p) A, &A_norm_inf);
	vector_norm_inf(NB, N, X, &X_norm_inf);
	vector_norm_inf(NB, N, B, &B_norm_inf);
	
	double resid1 = resid0 / (DBL_EPSILON * (A_norm_inf * X_norm_inf + B_norm_inf) * (double)N);
	
	if (resid1 >= 16.0) {
		fprintf(stderr, "Verification error: %g\n", resid1);
		return 1;
	}
	
	return 0;
}
