#ifndef integer
#error This benchmark requires integer to be defined to the equivalent Fortran integer type
#endif

#include <errno.h>
#include <float.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#define TYPE double
#include "../../utils/blas.h"
#include "../../utils/blas_init.h"
#include "../../utils/timer.h"

#if _OPENMP
#include "common/flat_init.h"
#else
#include "common/flat_init.h"
#endif
#include "../../utils/timer.h"

#if USE_SILLYMP
#include "sillymp.h"
#endif

static const integer ONE = 1;
static const integer MINUS_ONE = -1;
static const integer ZERO = 0;

static const double dzero = 0.0;
static const double done = 1.0;
static const double dminusone = -1.0;


void dgemm_wrapper(integer M, integer N, integer K, double ALPHA, double *A, double *B, double BETA, double *C) {
	lapack_dgemm("N", "N", &N, &M, &K, &ALPHA, B, &N, A, &K, &BETA, C, &N);
}


static inline double initializeElement(long I, long J, long i, long j) {
	static __thread uint64_t seed = 0;
	
	if ((i == 0) && (j == 0)) {
		seed = (I+i+1) * (J+j+1);
	}
	
	double result = (~0UL)/2UL;
	result -= seed;
	result /= ((~0UL / 2UL) + 1UL);
	seed = 6364136223846793005UL * seed + 1442695040888963407UL;
	
	return result;
}


static void dgemv_tile(integer M, integer N, double alpha, integer LDA, double const A[N][LDA], double const X[N], double beta, double Y[M]) {
	lapack_dgemv("N", &M, &N, &alpha, (const double *) A, &LDA, X, &ONE, &beta, Y, &ONE);
}


static void idamax_tile(integer N, const double X[N], double *result) {
	integer local_max = lapack_idamax(&N, X, &ONE);
	
	*result = max(*result, fabs(X[local_max-1]));
}


static void dasum_tile(integer M, integer N, const double X[M][N], double result[N]) {
	double local_result[M];
	
	for (integer j=0; j < M; j++)
		local_result[j] = 0.0;
	
	for (integer i=0; i < N; i++)
		for (integer j=0; j < M; j++)
			local_result[j] += fabs(X[j][i]);
	
	for (integer j=0; j < M; j++) {
		result[j] += local_result[j];
	}
}


static void substract(integer M, integer N, double a[M][N], double b[M][N], double c[M][N]) {
	integer total = M * N;
	
	lapack_dcopy(&total, a, &ONE, c, &ONE);
	lapack_daxpy(&total, &dminusone, b, &ONE, c, &ONE);
}


static void vector_norm_inf(integer N, double const V[N], double *result) {
	*result = 0.0;
	
	idamax_tile(N, V, result);
}


static void matrix_norm_inf(integer M, integer N, double const A[M][N], double *result) {
	typedef const double (*ac_p)[N];
	
	double *sums = malloc( ((long)M) * sizeof(double));
	
	for (long j=0; j < M; j++)
		sums[j] = 0.0;
	
	dasum_tile(M, N, A, sums);
	
	*result = 0.0;
	idamax_tile(M, sums, result);
	
	free(sums);
}


static void matrix_vector_product(integer M, integer N, double A[M][N], double X[N], double Y[M]) {
	lapack_dgemv("T", &N, &M, &done, (double const *) A, &N, X, &ONE, &dzero, Y, &ONE);
}


static void dump_matrix(integer M, integer N, double A[M][N]) {
	for (integer i=0; i < M ; i++) {
		for (integer j=0; j < N; j++) {
			printf("%g\t", A[i][j]);
		}
		printf("\n");
	}
}


int main(int argc, char *argv[])
{
	if (argc != 7) {
		fprintf(stderr, "Usage: %s <M> <N> <MBS> <NBS> <Index> <I|H|V|B|PW|SEQ>\n", argv[0]);
		exit(1);
	}
	
	unsigned long M = atol(argv[1]);
	unsigned long N = atol(argv[2]);
	unsigned long MBS = atol(argv[3]);
	unsigned long NBS = atol(argv[4]);
	unsigned long MaxBS = (MBS > NBS ? MBS : NBS);
	parseInitializationType(argv[6]);
	
	typedef double (*m_p)[N];
	typedef double const (*mc_p)[N];
	
	double (*A)[N] = (m_p) malloc(M * N * sizeof(double));
	double (*B)[N] = (m_p) malloc(N * N * sizeof(double));
	double (*C)[N] = (m_p) malloc(M * N * sizeof(double));
	
#if USE_SILLYMP
	sillymp_init(atoi(getenv("OMP_NUM_THREADS")));
#endif
	initializeUsingUserSpecification(M, N, MBS, NBS, A);
	initializeUsingUserSpecification(N, N, NBS, NBS, B);
#if USE_SILLYMP
	sillymp_end();
#endif
	force_blas_init();
	
	START_TIME();
	dgemm_wrapper(M, N, N, 1.0, (double *) A, (double *) B, 0.0, (double *) C);
	STOP_TIME();
	
#ifndef NO_CHECK
	//
	// Verification inspired in the HPCC DGEMM verification
	//
	double *X = malloc(1 * N * sizeof(double));
	double *Y = malloc(1 * M * sizeof(double));
	double *X2 = malloc(1 * N * sizeof(double));
	double *X3 = malloc(1 * M * sizeof(double));
	double *Z = malloc(1 * M * sizeof(double));
	
	double C_norm_inf;
	double X_norm_inf;
	
	typedef double (*n1_p)[N];
	typedef double const (*n1c_p)[N];
	typedef double (*m1_p)[M];
	typedef double const (*m1c_p)[M];
	
	initialize(1, N, 1, NBS, (n1_p) X);
	
	matrix_norm_inf(M, N, (mc_p) C, &C_norm_inf);
	vector_norm_inf(N, X, &X_norm_inf);
	
	matrix_vector_product(N, N, B, X, X2); // X2 := BX
	matrix_vector_product(M, N, A, X2, Y); // Y := A(X2)
	matrix_vector_product(M, N, C, X, Z); // Z := CX
	
	substract(1, M, // X3 := Y - Z === ABX - CX
		(m1_p) Y, (m1_p) Z, (m1_p) X);
	
	double resid0;
	vector_norm_inf(M, X3, &resid0);
	
	free(X);
	free(X2);
	free(X3);
	free(Y);
	free(Z);
	
	double scaled_residual = resid0 / (C_norm_inf * X_norm_inf * N * DBL_EPSILON);
	
	if (scaled_residual >= 16.0) {
		fprintf(stderr, "Verification error: %g\n", scaled_residual);
		return 1;
	}
#endif
	
	if (getenv("MEASURE_TIME") == NULL) {
		// Performance in MFLOPS
		printf ("%f", ((2.0 * (double) M * (double) N * (double) N)/GET_TIME()) );
	} else {
		printf("%f", GET_TIME() / 1000000.0);
	}
	
	return 0;
}

